from typing import Callable, List, Union

import numpy as np
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem


Molecule = Union[str, Chem.Mol]
FeaturesGenerator = Callable[[Molecule], np.ndarray]


FEATURES_GENERATOR_REGISTRY = {}


def register_features_generator(features_generator_name: str) -> Callable[[FeaturesGenerator], FeaturesGenerator]:
    """
    Creates a decorator which registers a features generator in a global dictionary to enable access by name.

    :param features_generator_name: The name to use to access the features generator.
    :return: A decorator which will add a features generator to the registry using the specified name.
    """
    def decorator(features_generator: FeaturesGenerator) -> FeaturesGenerator:
        FEATURES_GENERATOR_REGISTRY[features_generator_name] = features_generator
        return features_generator

    return decorator


def get_features_generator(features_generator_name: str) -> FeaturesGenerator:
    """
    Gets a registered features generator by name.

    :param features_generator_name: The name of the features generator.
    :return: The desired features generator.
    """
    if features_generator_name not in FEATURES_GENERATOR_REGISTRY:
        raise ValueError(f'Features generator "{features_generator_name}" could not be found. '
                         f'If this generator relies on rdkit features, you may need to install descriptastorus.')

    return FEATURES_GENERATOR_REGISTRY[features_generator_name]


def get_available_features_generators() -> List[str]:
    """Returns a list of names of available features generators."""
    return list(FEATURES_GENERATOR_REGISTRY.keys())


MORGAN_RADIUS = 2
MORGAN_NUM_BITS = 2048


@register_features_generator('morgan')
def morgan_binary_features_generator(mol: Molecule,
                                     radius: int = MORGAN_RADIUS,
                                     num_bits: int = MORGAN_NUM_BITS) -> np.ndarray:
    """
    Generates a binary Morgan fingerprint for a molecule.

    :param mol: A molecule (i.e., either a SMILES or an RDKit molecule).
    :param radius: Morgan fingerprint radius.
    :param num_bits: Number of bits in Morgan fingerprint.
    :return: A 1D numpy array containing the binary Morgan fingerprint.
    """
    mol = Chem.MolFromSmiles(mol) if type(mol) == str else mol
    features_vec = AllChem.GetMorganFingerprintAsBitVect(mol, radius, nBits=num_bits)
    features = np.zeros((1,))
    DataStructs.ConvertToNumpyArray(features_vec, features)

    return features


@register_features_generator('morgan_count')
def morgan_counts_features_generator(mol: Molecule,
                                     radius: int = MORGAN_RADIUS,
                                     num_bits: int = MORGAN_NUM_BITS) -> np.ndarray:
    """
    Generates a counts-based Morgan fingerprint for a molecule.

    :param mol: A molecule (i.e., either a SMILES or an RDKit molecule).
    :param radius: Morgan fingerprint radius.
    :param num_bits: Number of bits in Morgan fingerprint.
    :return: A 1D numpy array containing the counts-based Morgan fingerprint.
    """
    mol = Chem.MolFromSmiles(mol) if type(mol) == str else mol
    features_vec = AllChem.GetHashedMorganFingerprint(mol, radius, nBits=num_bits)
    features = np.zeros((1,))
    DataStructs.ConvertToNumpyArray(features_vec, features)

    return features


try:
    from descriptastorus.descriptors import rdDescriptors, rdNormalizedDescriptors

    @register_features_generator('rdkit_2d')
    def rdkit_2d_features_generator(mol: Molecule) -> np.ndarray:
        """
        Generates RDKit 2D features for a molecule.

        :param mol: A molecule (i.e., either a SMILES or an RDKit molecule).
        :return: A 1D numpy array containing the RDKit 2D features.
        """
        smiles = Chem.MolToSmiles(mol, isomericSmiles=True) if type(mol) != str else mol
        generator = rdDescriptors.RDKit2D()
        features = generator.process(smiles)[1:]

        return features

    @register_features_generator('rdkit_2d_normalized')
    def rdkit_2d_normalized_features_generator(mol: Molecule) -> np.ndarray:
        """
        Generates RDKit 2D normalized features for a molecule.

        :param mol: A molecule (i.e., either a SMILES or an RDKit molecule).
        :return: A 1D numpy array containing the RDKit 2D normalized features.
        """
        smiles = Chem.MolToSmiles(mol, isomericSmiles=True) if type(mol) != str else mol
        generator = rdNormalizedDescriptors.RDKit2DNormalized()
        features = generator.process(smiles)[1:]

        return features
except ImportError:
    @register_features_generator('rdkit_2d')
    def rdkit_2d_features_generator(mol: Molecule) -> np.ndarray:
        """Mock implementation raising an ImportError if descriptastorus cannot be imported."""
        raise ImportError('Failed to import descriptastorus. Please install descriptastorus '
                          '(https://github.com/bp-kelley/descriptastorus) to use RDKit 2D features.')

    @register_features_generator('rdkit_2d_normalized')
    def rdkit_2d_normalized_features_generator(mol: Molecule) -> np.ndarray:
        """Mock implementation raising an ImportError if descriptastorus cannot be imported."""
        raise ImportError('Failed to import descriptastorus. Please install descriptastorus '
                          '(https://github.com/bp-kelley/descriptastorus) to use RDKit 2D normalized features.')
except AttributeError as e:
    from scipy import __version__ as scipy_version
    from descriptastorus import __version__ as descriptastorus_version

    raise AttributeError('`descriptastorus==2.6.1` and `==2.5.1` are incompatible with `scipy<1.9`. Please try changing '
                    'descriptastorus versions or update scipy to avoid issues with `scipy.stats.gibrat`. '
                    f'Your versions are descriptastorus: {descriptastorus_version} and scipy {scipy_version}') from e

"""
Custom features generator template.

Note: The name you use to register the features generator is the name
you will specify on the command line when using the --features_generator <name> flag.
Ex. python train.py ... --features_generator custom ...

@register_features_generator('custom')
def custom_features_generator(mol: Molecule) -> np.ndarray:
    # If you want to use the SMILES string
    smiles = Chem.MolToSmiles(mol, isomericSmiles=True) if type(mol) != str else mol

    # If you want to use the RDKit molecule
    mol = Chem.MolFromSmiles(mol) if type(mol) == str else mol

    # Replace this with code which generates features from the molecule
    features = np.array([0, 0, 1])

    return features
"""



from .model import MoleculeModel
from .mpn import MPN, MPNEncoder
from .ffn import MultiReadout, FFNAtten

__all__ = [
    'MoleculeModel',
    'MPN',
    'MPNEncoder',
    'MultiReadout',
    'FFNAtten'
]



from logging import Logger
import os
import pickle
from typing import Dict, List, Union
from copy import deepcopy

import numpy as np
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.svm import SVC, SVR
from sklearn.linear_model import SGDClassifier, SGDRegressor
from tqdm import trange, tqdm

from chemprop.args import SklearnTrainArgs
from chemprop.data import MoleculeDataset, split_data, get_task_names, get_data
from chemprop.features import get_features_generator
from chemprop.train import cross_validate, evaluate_predictions
from chemprop.utils import save_smiles_splits


def predict(model: Union[RandomForestRegressor, RandomForestClassifier, SVR, SVC],
            model_type: str,
            dataset_type: str,
            features: List[np.ndarray]) -> List[List[float]]:
    """
    Predicts using a scikit-learn model.

    :param model: The trained scikit-learn model to make predictions with.
    :param model_type: The type of model.
    :param dataset_type: The type of dataset.
    :param features: The data features used as input for the model.
    :return: A list of lists of floats containing the predicted values.
    """
    if dataset_type == 'regression':
        preds = model.predict(features)

        if len(preds.shape) == 1:
            preds = [[pred] for pred in preds]
    elif dataset_type == 'classification':
        if model_type == 'random_forest':
            preds = model.predict_proba(features)

            if type(preds) == list:
                # Multiple tasks
                num_tasks, num_preds = len(preds), len(preds[0])
                preds = [[preds[i][j, 1] for i in range(num_tasks)] for j in range(num_preds)]
            else:
                # One task
                preds = [[preds[i, 1]] for i in range(len(preds))]
        elif model_type == 'svm':
            preds = model.decision_function(features)
            preds = [[pred] for pred in preds]
        else:
            raise ValueError(f'Model type "{model_type}" not supported')
    else:
        raise ValueError(f'Dataset type "{dataset_type}" not supported')

    return preds

def impute_sklearn(model: Union[RandomForestRegressor, RandomForestClassifier, SVR, SVC],
                   train_data: MoleculeDataset,
                   args: SklearnTrainArgs,
                   logger: Logger = None,
                   threshold: float = 0.5) -> List[float]:
    """
    Trains a single-task scikit-learn model, meaning a separate model is trained for each task.

    This is necessary if some tasks have None (unknown) values.

    :param model: The scikit-learn model to train.
    :param train_data: The training data.
    :param args: A :class:`~chemprop.args.SklearnTrainArgs` object containing arguments for
                 training the scikit-learn model.
    :param logger: A logger to record output.
    :param theshold: Threshold for classification tasks.
    :return: A list of list of target values.
    """
    num_tasks = train_data.num_tasks()
    new_targets=deepcopy(train_data.targets())
    
    if logger is not None:
        debug = logger.debug
    else:
        debug = print
        
    debug('Imputation')
    
    for task_num in trange(num_tasks):
        impute_train_features = [features for features, targets in zip(train_data.features(), train_data.targets()) if targets[task_num] is None]
        if len(impute_train_features) > 0:
            train_features, train_targets = zip(*[(features, targets[task_num])
                                              for features, targets in zip(train_data.features(), train_data.targets())
                                              if targets[task_num] is not None])
            if args.impute_mode == 'single_task':
                model.fit(train_features, train_targets)
                impute_train_preds = predict(
                    model=model,
                    model_type=args.model_type,
                    dataset_type=args.dataset_type,
                    features=impute_train_features
                )
                impute_train_preds = [pred[0] for pred in impute_train_preds]
            elif args.impute_mode == 'median' and args.dataset_type == 'regression':
                impute_train_preds = [np.median(train_targets)] * len(new_targets)
            elif args.impute_mode == 'mean' and args.dataset_type == 'regression':
                impute_train_preds = [np.mean(train_targets)] * len(new_targets)
            elif args.impute_mode == 'frequent' and args.dataset_type == 'classification':
                impute_train_preds = [np.argmax(np.bincount(train_targets))] * len(new_targets)
            elif args.impute_mode == 'linear' and args.dataset_type == 'regression':
                reg = SGDRegressor(alpha=0.01).fit(train_features, train_targets)
                impute_train_preds = reg.predict(impute_train_features)
            elif args.impute_mode == 'linear' and args.dataset_type == 'classification':
                cls = SGDClassifier().fit(train_features, train_targets)
                impute_train_preds = cls.predict(impute_train_features)
            else:
                raise ValueError("Invalid combination of imputation mode and dataset type.")   

            #Replace targets
            ctr = 0
            for i in range(len(new_targets)):
                if new_targets[i][task_num] is None:
                    value = impute_train_preds[ctr]
                    if args.dataset_type == 'classification':
                        value = int(value > threshold)
                    new_targets[i][task_num] = value
                    ctr += 1
                    
    return new_targets


def single_task_sklearn(model: Union[RandomForestRegressor, RandomForestClassifier, SVR, SVC],
                        train_data: MoleculeDataset,
                        test_data: MoleculeDataset,
                        metrics: List[str],
                        args: SklearnTrainArgs,
                        logger: Logger = None) -> List[float]:
    """
    Trains a single-task scikit-learn model, meaning a separate model is trained for each task.

    This is necessary if some tasks have None (unknown) values.

    :param model: The scikit-learn model to train.
    :param train_data: The training data.
    :param test_data: The test data.
    :param metrics: A list of names of metric functions.
    :param args: A :class:`~chemprop.args.SklearnTrainArgs` object containing arguments for
                 training the scikit-learn model.
    :param logger: A logger to record output.
    :return: A dictionary mapping each metric in :code:`metrics` to a list of values for each task.
    """
    scores = {}
    num_tasks = train_data.num_tasks()
    for task_num in trange(num_tasks):
        # Only get features and targets for molecules where target is not None
        train_features, train_targets = zip(*[(features, targets[task_num])
                                              for features, targets in zip(train_data.features(), train_data.targets())
                                              if targets[task_num] is not None])
        test_features, test_targets = zip(*[(features, targets[task_num])
                                            for features, targets in zip(test_data.features(), test_data.targets())
                                            if targets[task_num] is not None])

        model.fit(train_features, train_targets)

        test_preds = predict(
            model=model,
            model_type=args.model_type,
            dataset_type=args.dataset_type,
            features=test_features
        )
        test_targets = [[target] for target in test_targets]

        score = evaluate_predictions(
            preds=test_preds,
            targets=test_targets,
            num_tasks=1,
            metrics=metrics,
            dataset_type=args.dataset_type,
            logger=logger
        )
        for metric in metrics:
            if metric not in scores:
                scores[metric] = []
            scores[metric].append(score[metric][0])

    return scores


def multi_task_sklearn(model: Union[RandomForestRegressor, RandomForestClassifier, SVR, SVC],
                       train_data: MoleculeDataset,
                       test_data: MoleculeDataset,
                       metrics: List[str],
                       args: SklearnTrainArgs,
                       logger: Logger = None) -> Dict[str, List[float]]:
    """
    Trains a multi-task scikit-learn model, meaning one model is trained simultaneously on all tasks.

    This is only possible if none of the tasks have None (unknown) values.

    :param model: The scikit-learn model to train.
    :param train_data: The training data.
    :param test_data: The test data.
    :param metrics: A list of names of metric functions.
    :param args: A :class:`~chemprop.args.SklearnTrainArgs` object containing arguments for
                 training the scikit-learn model.
    :param logger: A logger to record output.
    :return: A dictionary mapping each metric in :code:`metrics` to a list of values for each task.
    """
    num_tasks = train_data.num_tasks()

    train_targets = train_data.targets()

    if args.impute_mode:
        train_targets = impute_sklearn(model=model,
                                       train_data=train_data,
                                       args=args,
                                       logger=logger)
    elif any(None in sublist for sublist in train_targets):
        raise ValueError("Missing target values not tolerated for multi-task sklearn models." 
                         "Use either --single_task to train multiple single-task models or impute"
                         " targets via --impute_mode  <model/linear/median/mean/frequent>.")
        
    if train_data.num_tasks() == 1:
        train_targets = [targets[0] for targets in train_targets]

    # Train
    model.fit(train_data.features(), train_targets)

    # Save model
    with open(os.path.join(args.save_dir, 'model.pkl'), 'wb') as f:
        pickle.dump(model, f)

    test_preds = predict(
        model=model,
        model_type=args.model_type,
        dataset_type=args.dataset_type,
        features=test_data.features()
    )

    scores = evaluate_predictions(
        preds=test_preds,
        targets=test_data.targets(),
        num_tasks=num_tasks,
        metrics=metrics,
        dataset_type=args.dataset_type,
        logger=logger
    )

    return scores


def run_sklearn(args: SklearnTrainArgs,
                data: MoleculeDataset,
                logger: Logger = None) -> Dict[str, List[float]]:
    """
    Loads data, trains a scikit-learn model, and returns test scores for the model checkpoint with the highest validation score.

    :param args: A :class:`~chemprop.args.SklearnTrainArgs` object containing arguments for
                 loading data and training the scikit-learn model.
    :param data: A :class:`~chemprop.data.MoleculeDataset` containing the data.
    :param logger: A logger to record output.
    :return: A dictionary mapping each metric in :code:`metrics` to a list of values for each task.
    """
    if logger is not None:
        debug, info = logger.debug, logger.info
    else:
        debug = info = print

    debug('Loading data')
    data = get_data(path=args.data_path,
                    features_path=args.features_path,
                    smiles_columns=args.smiles_columns,
                    target_columns=args.target_columns,
                    logger=logger)
    args.task_names = get_task_names(path=args.data_path,
                                     smiles_columns=args.smiles_columns,
                                     target_columns=args.target_columns,
                                     ignore_columns=args.ignore_columns)

    if args.model_type == 'svm' and data.num_tasks() != 1:
        raise ValueError(f'SVM can only handle single-task data but found {data.num_tasks()} tasks')

    debug(f'Splitting data with seed {args.seed}')
    # Need to have val set so that train and test sets are the same as when doing MPN
    train_data, _, test_data = split_data(
        data=data,
        split_type=args.split_type,
        seed=args.seed,
        sizes=args.split_sizes,
        num_folds=args.num_folds,
        args=args,
        logger=logger
    )

    if args.save_smiles_splits:
        save_smiles_splits(
            data_path=args.data_path,
            save_dir=args.save_dir,
            task_names=args.task_names,
            features_path=args.features_path,
            train_data=train_data,
            test_data=test_data,
            smiles_columns=args.smiles_columns,
            logger=logger
        )

    debug(f'Total size = {len(data):,} | train size = {len(train_data):,} | test size = {len(test_data):,}')

    debug('Computing morgan fingerprints')
    morgan_fingerprint = get_features_generator('morgan')
    for dataset in [train_data, test_data]:
        for datapoint in tqdm(dataset, total=len(dataset)):
            for s in datapoint.smiles:
                datapoint.extend_features(morgan_fingerprint(mol=s, radius=args.radius, num_bits=args.num_bits))

    debug('Building model')
    if args.dataset_type == 'regression':
        if args.model_type == 'random_forest':
            model = RandomForestRegressor(n_estimators=args.num_trees, n_jobs=-1, random_state=args.seed)
        elif args.model_type == 'svm':
            model = SVR()
        else:
            raise ValueError(f'Model type "{args.model_type}" not supported')
    elif args.dataset_type == 'classification':
        if args.model_type == 'random_forest':
            model = RandomForestClassifier(n_estimators=args.num_trees, n_jobs=-1, class_weight=args.class_weight, random_state=args.seed)
        elif args.model_type == 'svm':
            model = SVC()
        else:
            raise ValueError(f'Model type "{args.model_type}" not supported')
    else:
        raise ValueError(f'Dataset type "{args.dataset_type}" not supported')

    debug(model)

    model.train_args = args.as_dict()

    debug('Training')
    if args.single_task:
        scores = single_task_sklearn(
            model=model,
            train_data=train_data,
            test_data=test_data,
            metrics=args.metrics,
            args=args,
            logger=logger
        )
    else:
        scores = multi_task_sklearn(
            model=model,
            train_data=train_data,
            test_data=test_data,
            metrics=args.metrics,
            args=args,
            logger=logger
        )

    for metric in args.metrics:
        info(f'Test {metric} = {np.nanmean(scores[metric])}')

    return scores


def sklearn_train() -> None:
    """Parses scikit-learn training arguments and trains a scikit-learn model.

    This is the entry point for the command line command :code:`sklearn_train`.
    """
    cross_validate(args=SklearnTrainArgs().parse_args(), train_func=run_sklearn)



from abc import ABC, abstractmethod
from typing import Iterator, List

import numpy as np
from chemprop.data.data import MoleculeDataLoader
from scipy.special import erfinv, softmax, logit, expit
from scipy.optimize import fmin
from scipy.stats import t
from sklearn.isotonic import IsotonicRegression

from chemprop.data import MoleculeDataset, StandardScaler
from chemprop.models import MoleculeModel
from chemprop.uncertainty.uncertainty_predictor import build_uncertainty_predictor, UncertaintyPredictor
from chemprop.multitask_utils import reshape_values


class UncertaintyCalibrator(ABC):
    """
    Uncertainty calibrator class. Subclasses for each uncertainty calibration
    method. Subclasses should override the calibrate and apply functions for
    implemented metrics.
    """

    def __init__(
        self,
        uncertainty_method: str,
        calibration_method: str,
        interval_percentile: int,
        regression_calibrator_metric: str,
        calibration_data: MoleculeDataset,
        calibration_data_loader: MoleculeDataLoader,
        models: Iterator[MoleculeModel],
        scalers: Iterator[StandardScaler],
        num_models: int,
        dataset_type: str,
        loss_function: str,
        uncertainty_dropout_p: float,
        conformal_alpha: float,
        dropout_sampling_size: int,
        spectra_phase_mask: List[List[bool]],
    ):
        self.calibration_data = calibration_data
        self.calibration_data_loader = calibration_data_loader
        self.regression_calibrator_metric = regression_calibrator_metric
        self.interval_percentile = interval_percentile
        self.dataset_type = dataset_type
        self.uncertainty_method = uncertainty_method
        self.calibration_method = calibration_method
        self.loss_function = loss_function
        self.num_models = num_models
        self.conformal_alpha = conformal_alpha

        self.raise_argument_errors()

        self.calibration_predictor = build_uncertainty_predictor(
            test_data=calibration_data,
            test_data_loader=calibration_data_loader,
            models=models,
            scalers=scalers,
            num_models=num_models,
            dataset_type=dataset_type,
            loss_function=loss_function,
            uncertainty_method=uncertainty_method,
            uncertainty_dropout_p=uncertainty_dropout_p,
            conformal_alpha=conformal_alpha,
            dropout_sampling_size=dropout_sampling_size,
            individual_ensemble_predictions=False,
            spectra_phase_mask=spectra_phase_mask,
        )

        self.calibrate()

    @property
    @abstractmethod
    def label(self):
        """
        The string in saved results indicating the uncertainty method used.
        """

    def raise_argument_errors(self):
        """
        Raise errors for incompatibilities between dataset type and uncertainty method, or similar.
        """
        if self.dataset_type == "spectra":
            raise ValueError(
                "No uncertainty calibrators are implemented for the spectra dataset type."
            )
        if self.uncertainty_method in ["ensemble", "dropout"] and self.dataset_type in ["classification", "multiclass"]:
            raise ValueError(
                "Though ensemble and dropout uncertainty methods are available for classification \
                    multiclass dataset types, their outputs are not confidences and are not \
                    compatible with any implemented calibration methods for classification."
            )
        if self.uncertainty_method == "dirichlet":
            raise ValueError(
                "The Dirichlet uncertainty method returns an evidential uncertainty value rather than a \
                    class confidence. It is not compatible with any implemented calibration methods. \
                    To calibrate a model trained using the Dirichlet loss function, \
                    use the classification uncertainty method."
            )

    @abstractmethod
    def calibrate(self):
        """
        Fit calibration method for the calibration data.
        """

    @abstractmethod
    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        """
        Take in predictions and uncertainty parameters from a model and apply the calibration method using fitted parameters.
        """

    @abstractmethod
    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ) -> List[float]:
        """
        Takes in calibrated predictions and uncertainty parameters and returns the log probability density of that result.
        """


class ZScalingCalibrator(UncertaintyCalibrator):
    """
    A class that calibrates regression uncertainty models by applying
    a scaling value to the uncalibrated standard deviation, fitted by minimizing the
    negative log likelihood of a normal distribution around each prediction
    with scaling given by the uncalibrated variance. Method is described
    in https://arxiv.org/abs/1905.11659.
    """

    @property
    def label(self):
        if self.regression_calibrator_metric == "stdev":
            label = f"{self.uncertainty_method}_zscaling_stdev"
        else:  # interval
            label = (
                f"{self.uncertainty_method}_zscaling_{self.interval_percentile}interval"
            )
        return label

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError("Z Score Scaling is only compatible with regression datasets.")

    def calibrate(self):
        uncal_preds = np.array(self.calibration_predictor.get_uncal_preds())  # shape(data, tasks)
        uncal_vars = np.array(self.calibration_predictor.get_uncal_vars())
        targets = np.array(self.calibration_data.targets())
        mask = np.array(self.calibration_data.mask())
        self.num_tasks = len(mask)
        if self.calibration_data.is_atom_bond_targets:
            uncal_preds = [np.concatenate(x) for x in zip(*uncal_preds)]
            uncal_vars = [np.concatenate(x) for x in zip(*uncal_vars)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncal_preds = np.array(list(zip(*uncal_preds)))
            uncal_vars = np.array(list(zip(*uncal_vars)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        self.scaling = np.zeros(self.num_tasks)

        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_preds = uncal_preds[i][task_mask]
            task_vars = uncal_vars[i][task_mask]
            task_errors = task_preds - task_targets
            task_zscore = task_errors / np.sqrt(task_vars)

            def objective(scaler_value: float):
                scaled_vars = task_vars * scaler_value**2
                nll = np.log(2 * np.pi * scaled_vars) / 2 + (task_errors) ** 2 / (2 * scaled_vars)
                return nll.sum()

            initial_guess = np.std(task_zscore)
            sol = fmin(objective, initial_guess)

            if self.regression_calibrator_metric == "stdev":
                self.scaling[i] = sol
            else:  # interval
                self.scaling[i] = sol * erfinv(self.interval_percentile / 100) * np.sqrt(2)

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())
        uncal_vars = np.array(uncal_predictor.get_uncal_vars())
        if self.calibration_data.is_atom_bond_targets:
            cal_stdev = []
            sqrt_uncal_vars = [
                [np.sqrt(var) for var in uncal_var] for uncal_var in uncal_vars
            ]
            for sqrt_uncal_var in sqrt_uncal_vars:
                scaled_stdev = [var * s for var, s in zip(sqrt_uncal_var, self.scaling)]
                cal_stdev.append(scaled_stdev)
            return uncal_preds, cal_stdev
        else:
            cal_stdev = np.sqrt(uncal_vars) * np.expand_dims(self.scaling, axis=0)
            return uncal_preds.tolist(), cal_stdev.tolist()

    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ):
        unc_var = np.square(unc)
        preds = np.array(preds)
        targets = np.array(targets)
        mask = np.array(mask)
        if self.calibration_data.is_atom_bond_targets:
            unc_var = [np.concatenate(x) for x in zip(*unc_var)]
            preds = [np.concatenate(x) for x in zip(*preds)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            unc_var = np.array(list(zip(*unc_var)))
            preds = np.array(list(zip(*preds)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        nll = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_preds = preds[i][task_mask]
            task_targets = targets[i][task_mask]
            task_unc = unc_var[i][task_mask]
            task_nll = (
                np.log(2 * np.pi * task_unc) / 2
                + (task_preds - task_targets) ** 2 / (2 * task_unc)
            )
            nll.append(task_nll.mean())
        return nll


class TScalingCalibrator(UncertaintyCalibrator):
    """
    A class that calibrates regression uncertainty models using a variation of the
    ZScaling method. Instead, this method assumes that error is dominated by
    variance error as represented by the variance of the ensemble predictions.
    The scaling value is obtained by minimizing the negative log likelihood
    of the t distribution, including reductio term due to the number of ensemble models sampled.
    """

    @property
    def label(self):
        if self.regression_calibrator_metric == "stdev":
            label = f"{self.uncertainty_method}_tscaling_stdev"
        else:  # interval
            label = (
                f"{self.uncertainty_method}_tscaling_{self.interval_percentile}interval"
            )
        return label

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError("T Score Scaling is only compatible with regression datasets.")
        if self.uncertainty_method == "dropout":
            raise ValueError("T scaling not enabled with dropout variance uncertainty method.")
        if self.num_models == 1:
            raise ValueError("T scaling is intended for use with ensemble models.")

    def calibrate(self):
        uncal_preds = np.array(self.calibration_predictor.get_uncal_preds())  # shape(data, tasks)
        uncal_vars = np.array(self.calibration_predictor.get_uncal_vars())
        targets = np.array(self.calibration_data.targets())
        mask = np.array(self.calibration_data.mask())
        self.num_tasks = len(mask)
        if self.calibration_data.is_atom_bond_targets:
            uncal_preds = [np.concatenate(x) for x in zip(*uncal_preds)]
            uncal_vars = [np.concatenate(x) for x in zip(*uncal_vars)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncal_preds = np.array(list(zip(*uncal_preds)))
            uncal_vars = np.array(list(zip(*uncal_vars)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        self.scaling = np.zeros(self.num_tasks)

        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_preds = uncal_preds[i][task_mask]
            task_vars = uncal_vars[i][task_mask]
            std_error_of_mean = np.sqrt(
                task_vars / (self.num_models - 1)
            )  # reduced for number of samples and include Bessel's correction
            task_errors = task_preds - task_targets
            task_tscore = task_errors / std_error_of_mean

            def objective(scaler_value: np.ndarray):
                scaled_std = std_error_of_mean * scaler_value
                likelihood = t.pdf(
                    x=task_errors, df=self.num_models - 1, scale=scaled_std
                )  # scipy t distribution pdf
                nll = -1 * np.sum(np.log(likelihood), axis=0)
                return nll

            initial_guess = np.std(task_tscore)
            stdev_scaling = fmin(objective, initial_guess)
            if self.regression_calibrator_metric == "stdev":
                self.scaling[i] = stdev_scaling
            else:  # interval
                interval_scaling = stdev_scaling * t.ppf(
                    (self.interval_percentile / 100 + 1) / 2, df=self.num_models - 1
                )
                self.scaling[i] = interval_scaling

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())
        uncal_vars = np.array(uncal_predictor.get_uncal_vars())
        if self.calibration_data.is_atom_bond_targets:
            cal_stdev = []
            sqrt_uncal_vars = [
                [np.sqrt(var / (self.num_models - 1)) for var in uncal_var]
                for uncal_var in uncal_vars
            ]
            for sqrt_uncal_var in sqrt_uncal_vars:
                scaled_stdev = [var * s for var, s in zip(sqrt_uncal_var, self.scaling)]
                cal_stdev.append(scaled_stdev)
            return uncal_preds, cal_stdev
        else:
            cal_stdev = np.sqrt(uncal_vars / (self.num_models - 1)) * np.expand_dims(
                self.scaling, axis=0
            )
        return uncal_preds.tolist(), cal_stdev.tolist()

    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ):
        unc = np.square(unc)
        preds = np.array(preds)
        targets = np.array(targets)
        mask = np.array(mask)
        if self.calibration_data.is_atom_bond_targets:
            unc = [np.concatenate(x) for x in zip(*unc)]
            preds = [np.concatenate(x) for x in zip(*preds)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            unc = np.array(list(zip(*unc)))
            preds = np.array(list(zip(*preds)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        nll = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_preds = preds[i][task_mask]
            task_targets = targets[i][task_mask]
            task_unc = unc[i][task_mask]
            task_nll = -1 * t.logpdf(
                x=task_preds - task_targets, scale=task_unc, df=self.num_models - 1
            )
            nll.append(task_nll.mean())
        return nll


class ZelikmanCalibrator(UncertaintyCalibrator):
    """
    A calibrator for regression datasets that does not depend on a particular probability
    function form. Designed to be used with interval output. Uses the "CRUDE" method as
    described in https://arxiv.org/abs/2005.12496. As implemented here, the interval
    bounds are constrained to be symmetrical, though this is not required in the source method.
    The probability density to be used for NLL evaluator for the zelikman interval method is
    approximated here as a histogram function.
    """

    @property
    def label(self):
        if self.regression_calibrator_metric == "stdev":
            label = f"{self.uncertainty_method}_zelikman_stdev"
        else:
            label = f"{self.uncertainty_method}_zelikman_{self.interval_percentile}interval"
        return label

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError("Crude Scaling is only compatible with regression datasets.")

    def calibrate(self):
        uncal_preds = np.array(self.calibration_predictor.get_uncal_preds())  # shape(data, tasks)
        uncal_vars = np.array(self.calibration_predictor.get_uncal_vars())
        targets = np.array(self.calibration_data.targets())
        mask = np.array(self.calibration_data.mask())
        self.num_tasks = len(mask)
        if self.calibration_data.is_atom_bond_targets:
            uncal_preds = [np.concatenate(x) for x in zip(*uncal_preds)]
            uncal_vars = [np.concatenate(x) for x in zip(*uncal_vars)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncal_preds = np.array(list(zip(*uncal_preds)))
            uncal_vars = np.array(list(zip(*uncal_vars)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        self.histogram_parameters = []
        self.scaling = np.zeros(self.num_tasks)
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_preds = uncal_preds[i][task_mask]
            task_targets = targets[i][task_mask]
            task_vars = uncal_vars[i][task_mask]
            task_preds = np.abs(task_preds - task_targets) / np.sqrt(task_vars)
            if self.regression_calibrator_metric == "interval":
                interval_scaling = np.percentile(task_preds, self.interval_percentile)
                self.scaling[i] = interval_scaling
            else:
                symmetric_z = np.concatenate([task_preds, -1 * task_preds])
                std_scaling = np.std(symmetric_z, axis=0)
                self.scaling[i] = std_scaling
            # histogram parameters for nll calculation
            h_params = np.histogram(task_preds, bins="auto", density=True)
            self.histogram_parameters.append(h_params)

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())
        uncal_vars = np.array(uncal_predictor.get_uncal_vars())
        if self.calibration_data.is_atom_bond_targets:
            cal_stdev = []
            sqrt_uncal_vars = [
                [np.sqrt(var) for var in uncal_var] for uncal_var in uncal_vars
            ]
            for sqrt_uncal_var in sqrt_uncal_vars:
                scaled_stdev = [var * s for var, s in zip(sqrt_uncal_var, self.scaling)]
                cal_stdev.append(scaled_stdev)
            return uncal_preds, cal_stdev
        else:
            cal_stdev = np.sqrt(uncal_vars) * np.expand_dims(self.scaling, axis=0)
            return uncal_preds.tolist(), cal_stdev.tolist()

    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ):
        preds = np.array(preds)
        unc = np.array(unc)
        targets = np.array(targets)
        mask = np.array(mask)
        if self.calibration_data.is_atom_bond_targets:
            preds = [np.concatenate(x) for x in zip(*preds)]
            unc = [np.concatenate(x) for x in zip(*unc)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            preds = np.array(list(zip(*preds)))
            unc = np.array(list(zip(*unc)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        nll = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_preds = preds[i][task_mask]
            task_targets = targets[i][task_mask]
            task_stdev = unc[i][task_mask] / self.scaling[i]
            task_abs_z = np.abs(task_preds - task_targets) / task_stdev
            bin_edges = self.histogram_parameters[i][1]
            bin_magnitudes = self.histogram_parameters[i][0]
            bin_magnitudes = np.insert(bin_magnitudes, [0, len(bin_magnitudes)], 0)
            pred_bins = np.searchsorted(bin_edges, task_abs_z)
            # magnitude adjusted by stdev scale of the distribution and symmetry assumption
            task_likelihood = bin_magnitudes[pred_bins] / task_stdev / 2
            task_nll = -1 * np.log(task_likelihood)
            nll.append(task_nll.mean())
        return nll


class MVEWeightingCalibrator(UncertaintyCalibrator):
    """
    A method of calibration for models that have ensembles of individual models that
    make variance predictions. Minimizes the negative log likelihood for the
    predictions versus the targets by applying a weighted average across the
    variance predictions of the ensemble. Discussed in https://doi.org/10.1186/s13321-021-00551-x.
    """

    @property
    def label(self):
        if self.regression_calibrator_metric == "stdev":
            label = f"{self.uncertainty_method}_mve_weighting_stdev"
        else:  # interval
            label = f"{self.uncertainty_method}_mve_weighting_{self.interval_percentile}interval"
        return label

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError(
                f"MVE Weighting is only compatible with regression datasets! got: {self.dataset_type}"
            )
        if self.loss_function not in ["mve", "evidential"]:
            raise ValueError(
                "MVE Weighting calibration can only be carried out with MVE or Evidential loss function models."
            )
        if self.num_models == 1:
            raise ValueError(
                "MVE Weighting is only useful when weighting between results in an ensemble. Only one model was provided."
            )

    def calibrate(self):
        uncal_preds = np.array(self.calibration_predictor.get_uncal_preds())  # shape(data, tasks)
        individual_vars = np.array(
            self.calibration_predictor.get_individual_vars()
        )  # shape(models, data, tasks)
        targets = np.array(self.calibration_data.targets())
        mask = np.array(self.calibration_data.mask())
        self.num_tasks = len(mask)
        if self.calibration_data.is_atom_bond_targets:
            uncal_preds = [np.concatenate(x) for x in zip(*uncal_preds)]
            individual_vars = [np.array([np.concatenate(individual_vars[j][i][:, :]) for j in range(self.num_models)]) for i in range(self.num_tasks)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncal_preds = np.array(list(zip(*uncal_preds)))
            individual_vars = [individual_vars[:, :, i] for i in range(self.num_tasks)]
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        self.var_weighting = np.zeros([self.num_models, self.num_tasks])  # shape(models, tasks)

        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_preds = uncal_preds[i][task_mask]
            task_ind_vars = individual_vars[i][:, task_mask]
            task_errors = task_preds - task_targets

            def objective(scaler_values: np.ndarray):
                scaler_values = np.reshape(softmax(scaler_values), [-1, 1])  # (models, 1)
                scaled_vars = np.sum(
                    task_ind_vars * scaler_values, axis=0, keepdims=False
                )  # shape(data)
                nll = np.log(2 * np.pi * scaled_vars) / 2 + (task_errors) ** 2 / (2 * scaled_vars)
                nll = np.sum(nll)
                return nll

            initial_guess = np.ones(self.num_models)
            sol = fmin(objective, initial_guess)
            self.var_weighting[:, i] = softmax(sol)
        if self.regression_calibrator_metric == "stdev":
            self.scaling = np.repeat(1, self.num_tasks)
        else:  # interval
            self.scaling = np.repeat(erfinv(self.interval_percentile / 100) * np.sqrt(2), self.num_tasks)

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())
        uncal_individual_vars = np.array(uncal_predictor.get_individual_vars())
        weighted_vars = None
        for ind_vars, s in zip(uncal_individual_vars, self.var_weighting):
            if weighted_vars is None:
                weighted_vars = ind_vars
                for i in range(len(s)):
                    weighted_vars[i] *= s[i]
            else:
                for i in range(len(s)):
                    weighted_vars[i] += ind_vars[i] * s[i]
        if self.calibration_data.is_atom_bond_targets:
            sqrt_weighted_vars = [np.array([np.sqrt(var) for var in uncal_var]) for uncal_var in weighted_vars]
            weighted_stdev = sqrt_weighted_vars * self.scaling
            natom_targets = len(self.calibration_data[0].atom_targets) if self.calibration_data[0].atom_targets is not None else 0
            nbond_targets = len(self.calibration_data[0].bond_targets) if self.calibration_data[0].bond_targets is not None else 0
            weighted_stdev = reshape_values(
                weighted_stdev,
                self.calibration_data,
                natom_targets,
                nbond_targets,
            )
            return uncal_preds, weighted_stdev
        else:
            weighted_stdev = np.sqrt(weighted_vars) * self.scaling
            return uncal_preds.tolist(), weighted_stdev.tolist()

    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ):
        unc_var = np.square(unc)
        preds = np.array(preds)
        targets = np.array(targets)
        mask = np.array(mask)
        if self.calibration_data.is_atom_bond_targets:
            unc_var = [np.concatenate(np.square(x)) for x in zip(*unc)]
            preds = [np.concatenate(x) for x in zip(*preds)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            unc_var = np.array(list(zip(*unc_var)))
            preds = np.array(list(zip(*preds)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        nll = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_preds = preds[i][task_mask]
            task_targets = targets[i][task_mask]
            task_unc = unc_var[i][task_mask]
            task_nll = (
                np.log(2 * np.pi * task_unc) / 2
                + (task_preds - task_targets) ** 2 / (2 * task_unc)
            )
            nll.append(task_nll.mean())
        return nll


class PlattCalibrator(UncertaintyCalibrator):
    """
    A calibration method for classification datasets based on the Platt scaling algorithm.
    As discussed in https://arxiv.org/abs/1706.04599.
    """

    @property
    def label(self):
        return f"{self.uncertainty_method}_platt_confidence"

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "classification":
            raise ValueError("Platt scaling is only implemented for classification dataset types.")

    def calibrate(self):
        uncal_preds = np.array(
            self.calibration_predictor.get_uncal_preds()
        )  # shape(data, tasks)
        targets = np.array(self.calibration_data.targets())
        if self.calibration_data.is_atom_bond_targets:
            uncal_preds = [np.concatenate(x) for x in zip(*uncal_preds)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncal_preds = np.array(list(zip(*uncal_preds)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
        mask = np.array(self.calibration_data.mask())
        self.num_tasks = len(mask)
        # If train class sizes are available, set Bayes corrected calibration targets
        if self.calibration_predictor.train_class_sizes is not None:
            class_size_correction = True
            train_class_sizes = np.sum(
                self.calibration_predictor.train_class_sizes, axis=0
            )  # shape(tasks, 2)
            negative_target = 1 / (train_class_sizes[:, 0] + 2)
            positive_target = (train_class_sizes[:, 1] + 1) / (train_class_sizes[:, 1] + 2)
            print(
                "Platt scaling for calibration uses Bayesian correction against training set overfitting, "
                "replacing calibration targets [0,1] with adjusted values."
            )
        else:
            class_size_correction = False
            print(
                "Class sizes used in training models unavailable in checkpoints before Chemprop v1.5.0. "
                "No Bayesian correction perfomed as part of class scaling."
            )

        platt_parameters = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_preds = uncal_preds[i][task_mask]
            if class_size_correction:
                task_targets[task_targets == 0] = negative_target[i]
                task_targets[task_targets == 1] = positive_target[i]
                print(
                    f"Platt Bayesian correction for task {i} in calibration replacing [0,1] targets with {[negative_target[i], positive_target[i]]}"
                )

            def objective(parameters: np.ndarray):
                a = parameters[0]
                b = parameters[1]
                scaled_preds = expit(a * logit(task_preds) + b)
                nll = -1 * np.sum(
                    task_targets * np.log(scaled_preds)
                    + (1 - task_targets) * np.log(1 - scaled_preds)
                )
                return nll

            initial_guess = [1, 0]
            sol = fmin(objective, initial_guess)
            platt_parameters.append(sol)

        platt_parameters = np.array(platt_parameters)  # shape(task, 2)
        self.platt_a = platt_parameters[:, 0]
        self.platt_b = platt_parameters[:, 1]

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())  # shape(data, task)
        cal_preds = expit(
            np.expand_dims(self.platt_a, axis=0) * logit(uncal_preds)
            + np.expand_dims(self.platt_b, axis=0)
        )
        return uncal_preds.tolist(), cal_preds.tolist()

    def nll(self, preds: List[List[float]], unc: List[List[float]], targets: List[List[float]], mask: List[List[bool]]):
        targets = np.array(targets)
        unc = np.array(unc)
        mask = np.array(mask)
        if self.calibration_data.is_atom_bond_targets:
            targets = [np.concatenate(x) for x in zip(*targets)]
            unc = [np.concatenate(x) for x in zip(*unc)]
        else:
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
            unc = np.array(list(zip(*unc)))
        nll = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_unc = unc[i][task_mask]

            likelihood = task_unc * task_targets + (1 - task_unc) * (1 - task_targets)
            task_nll = -1 * np.log(likelihood)
            nll.append(task_nll.mean())
        return nll


class IsotonicCalibrator(UncertaintyCalibrator):
    """
    A calibration method for classification datasets based on the isotonic regression algorithm.
    In effect, the method transforms incoming uncalibrated confidences using a histogram-like
    function where the range of each transforming bin and its magnitude is learned.
    As discussed in https://arxiv.org/abs/1706.04599.
    """

    @property
    def label(self):
        return f"{self.uncertainty_method}_isotonic_confidence"

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "classification":
            raise ValueError(
                "Isotonic Regression is only implemented for classification dataset types."
            )

    def calibrate(self):
        uncal_preds = np.array(
            self.calibration_predictor.get_uncal_preds()
        )  # shape(data, tasks)
        targets = np.array(self.calibration_data.targets())
        mask = np.array(self.calibration_data.mask())
        self.num_tasks = len(mask)

        if self.calibration_data.is_atom_bond_targets:
            uncal_preds = [np.concatenate(x) for x in zip(*uncal_preds)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncal_preds = np.array(list(zip(*uncal_preds)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))

        isotonic_models = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_preds = uncal_preds[i][task_mask]

            isotonic_model = IsotonicRegression(y_min=0, y_max=1, out_of_bounds="clip")
            isotonic_model.fit(task_preds, task_targets)
            isotonic_models.append(isotonic_model)

        self.isotonic_models = isotonic_models

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())  # shape(data, task)
        if self.calibration_data.is_atom_bond_targets:
            cal_preds = []
            uncal_preds_list = [np.concatenate(x) for x in zip(*uncal_preds)]
            for i, iso_model in enumerate(self.isotonic_models):
                task_preds = uncal_preds_list[i]
                task_cal = iso_model.predict(task_preds)
                transpose_cal_preds = [task_cal]
                cal_preds.append(np.transpose(transpose_cal_preds))
            return uncal_preds, cal_preds
        else:
            transpose_cal_preds = []
            for i, iso_model in enumerate(self.isotonic_models):
                task_preds = uncal_preds[:, i]
                task_cal = iso_model.predict(task_preds)
                transpose_cal_preds.append(task_cal)
            cal_preds = np.transpose(transpose_cal_preds)
            return uncal_preds.tolist(), cal_preds.tolist()

    def nll(self, preds: List[List[float]], unc: List[List[float]], targets: List[List[float]], mask: List[List[bool]]):
        targets = np.array(targets)
        mask = np.array(mask)
        unc = np.array(unc)
        if self.calibration_data.is_atom_bond_targets:
            targets = [np.concatenate(x) for x in zip(*targets)]
            unc = [np.concatenate(x) for x in zip(*unc)]
        else:
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))
            unc = np.array(list(zip(*unc)))
        nll = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_unc = unc[i][task_mask]

            likelihood = task_unc * task_targets + (1 - task_unc) * (1 - task_targets)
            task_nll = -1 * np.log(likelihood)
            nll.append(task_nll.mean())
        return nll


class IsotonicMulticlassCalibrator(UncertaintyCalibrator):
    """
    A multiclass method for classification datasets based on the isotonic regression algorithm.
    In effect, the method transforms incoming uncalibrated confidences using a histogram-like
    function where the range of each transforming bin and its magnitude is learned. Uses a
    one-against-all aggregation scheme for convertering between binary and multiclass classifiers.
    As discussed in https://arxiv.org/abs/1706.04599.
    """

    @property
    def label(self):
        return f"{self.uncertainty_method}_isotonic_confidence"

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "multiclass":
            raise ValueError(
                "Isotonic Multiclass Regression is only implemented for multiclass dataset types."
            )

    def calibrate(self):
        uncal_preds = np.array(
            self.calibration_predictor.get_uncal_preds()
        )  # shape(data, tasks, num_classes)
        targets = np.array(self.calibration_data.targets(), dtype=float)  # shape(data, tasks)
        mask = np.array(self.calibration_data.mask())
        self.num_tasks = len(mask)
        self.num_classes = uncal_preds.shape[2]

        isotonic_models = []
        for i in range(self.num_tasks):
            isotonic_models.append([])
            task_mask = mask[i]
            task_targets = targets[task_mask, i]  # shape(data)
            task_preds = uncal_preds[task_mask, i]
            for j in range(self.num_classes):
                class_preds = task_preds[:, j]  # shape(data)
                positive_class_targets = task_targets == j

                class_targets = np.ones_like(class_preds)
                class_targets[positive_class_targets] = 1
                class_targets[~positive_class_targets] = 0

                isotonic_model = IsotonicRegression(y_min=0, y_max=1, out_of_bounds="clip")
                isotonic_model.fit(class_preds, class_targets)
                isotonic_models[i].append(isotonic_model)

        self.isotonic_models = isotonic_models  # shape(tasks, classes)

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())  # shape(data, task, class)
        transpose_cal_preds = []
        for i in range(self.num_tasks):
            transpose_cal_preds.append([])
            for j in range(self.num_classes):
                class_preds = uncal_preds[:, i, j]
                class_cal = self.isotonic_models[i][j].predict(class_preds)
                transpose_cal_preds[i].append(class_cal)  # shape (task, class, data)
        cal_preds = np.transpose(transpose_cal_preds, [2, 0, 1])  # shape(data, task, class)
        cal_preds = cal_preds / np.sum(cal_preds, axis=2, keepdims=True)
        return uncal_preds.tolist(), cal_preds.tolist()

    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ):
        targets = np.array(targets, dtype=int)  # shape(data, tasks)
        mask = np.array(mask)
        unc = np.array(unc)
        preds = np.array(preds)
        nll = []
        for i in range(targets.shape[1]):
            task_mask = mask[i]
            task_preds = unc[task_mask, i]
            task_targets = targets[task_mask, i]  # shape(data)
            bin_targets = np.zeros_like(preds[:, 0, :])  # shape(data, classes)
            bin_targets[np.arange(targets.shape[0]), task_targets] = 1
            task_likelihood = np.sum(bin_targets * task_preds, axis=1)
            task_nll = -1 * np.log(task_likelihood)
            nll.append(task_nll.mean())
        return nll


class ConformalMulticlassCalibrator(UncertaintyCalibrator):
    """
    Conformal Calibrator. Outputs binary values for whether each class should be included in the
    conformal set for each task.
    As discussed in https://arxiv.org/abs/2107.07511.
    """

    @property
    def label(self):
        return f"conformal_{self.conformal_alpha}"

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "multiclass":
            raise ValueError("Conformal is only implemented for multiclass dataset types.")

    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ):
        raise NotImplementedError(
            "The NLL uncertainty evaluation method for classification models has not been implemented for use with the conformal classification calibration method."
            )

    @staticmethod
    def nonconformity_scores(uncal_preds):
        """Fixed per class. Example is for basic conformal.

        Args:
            uncal_preds (torch.Tensor): a tensor of shape `n x t x c`, where `n`
            is the number of examples, `t` the number of tasks, and `c` the number
            of classes, containing the uncalibrated model predictions.
        
        Returns:
            scores (torch.Tensor): [num_examples, num_tasks, num_classes]
        """
        return -uncal_preds

    def calibrate(self):
        uncal_preds = np.array(
            self.calibration_predictor.get_uncal_preds()
        )  # shape(data, tasks, num_classes)
        targets = np.array(self.calibration_data.targets(), dtype=float)  # shape(data, tasks)
        mask = np.array(self.calibration_data.mask(), dtype=bool)
        num_data, self.num_tasks, self.num_classes = uncal_preds.shape

        all_scores = self.nonconformity_scores(uncal_preds)
        self.qhats = []

        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_scores = np.take_along_axis(
                all_scores[task_mask, i], targets[task_mask, i].reshape(-1, 1).astype(int), axis=1
            ).squeeze(1)  # shape(valid_data)
            q_level = np.ceil((num_data + 1) * (1 - self.conformal_alpha)) / num_data
            qhat = np.quantile(task_scores, q_level, interpolation='higher')
            self.qhats.append(qhat)

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())  # shape(data, task)
        cal_preds = np.zeros_like(uncal_preds, dtype=int)
        all_scores = self.nonconformity_scores(uncal_preds)
        for task_id, qhat in enumerate(self.qhats):
            cal_preds[:, task_id] = all_scores[:, task_id] <= qhat
        return uncal_preds.tolist(), cal_preds.tolist()


class ConformalAdaptiveMulticlassCalibrator(ConformalMulticlassCalibrator):
    """
    Adaptive Conformal Calibrator. Outputs binary values for whether each class should be
    included in the conformal set for each task.
    As discussed in https://arxiv.org/abs/2107.07511.
    """

    @property
    def label(self):
        return f"conformal_adaptive_{self.conformal_alpha}"

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "multiclass":
            raise ValueError("Conformal Adaptive is only implemented for multiclass dataset types.")

    @staticmethod
    def nonconformity_scores(uncal_preds):
        """Fixed per class. Example is for adaptive conformal.

        Args:
            uncal_preds (torch.Tensor): a tensor of shape `n x t x c`, where `n`
            is the number of examples, `t` the number of tasks, and `c` the number
            of classes, containing the uncalibrated model predictions.
        
        Returns:
            scores (torch.Tensor): [num_examples, num_tasks, num_classes]
        """
        sort_inds = np.argsort(-uncal_preds, axis=2)
        sorted_preds = np.take_along_axis(uncal_preds, sort_inds, axis=2)
        sorted_scores = sorted_preds.cumsum(axis=2)
        unsort_inds = np.argsort(sort_inds, axis=2)
        unsorted_scores = np.take_along_axis(sorted_scores, unsort_inds, axis=2)
        return unsorted_scores


class ConformalMultilabelCalibrator(UncertaintyCalibrator):
    """
    Conformal Calibrator for Multilabel datasets. Creates conformal in-set and conformal out-set such that
    for 1-alpha proportion of datapoints, the set of labels is bounded by the in-set and out-set. That is,
    the conformal in-set is contained in the set of actual labels and the set of actual labels is contained
    in the conformal out-set.
    As discussed in https://arxiv.org/abs/2004.10181.
    """

    @property
    def label(self):
        return f"conformal_multilabel_{self.conformal_alpha}"

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "classification":
            raise ValueError("Conformal is only implemented for classification dataset types.")

    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ):
        raise NotImplementedError(
            "The NLL uncertainty evaluation method for classification models has not been implemented for use with the conformal classification calibration method."
            )

    @staticmethod
    def nonconformity_scores(uncal_preds):
        """Fixed per class. Example is for multilabel conformal.

        Args:
            uncal_preds (torch.Tensor): a tensor of shape `n x t`, where `n`
            is the number of examples, and `t` the number of tasks,
            containing the uncalibrated model predictions.
        
        Returns:
            scores (torch.Tensor): [num_examples, num_tasks]
        """
        return -uncal_preds

    def calibrate(self):
        uncal_preds = np.array(
            self.calibration_predictor.get_uncal_preds()
        )  # shape(data, tasks)
        targets = np.array(self.calibration_data.targets(), dtype=bool)  # shape(data, tasks)
        mask = np.array(self.calibration_data.mask(), dtype=bool)
        self.num_data, self.num_tasks = targets.shape

        has_zeros = np.any(targets == 0, axis=1)
        inds_zeros = targets[has_zeros] == 0
        scores_in = self.nonconformity_scores(uncal_preds[has_zeros])
        masked_scores_in = scores_in * inds_zeros + np.nan_to_num(
            np.inf * (1 - inds_zeros).astype(float)
        )
        calibration_scores_in = np.empty(masked_scores_in.shape[0])
        for i, score_row in enumerate(masked_scores_in):
            data_mask = mask.T[i]
            masked_scores = score_row[data_mask]
            calibration_scores_in[i] = np.min(masked_scores)

        has_ones = np.any(targets == 1, axis=1)
        inds_ones = targets[has_ones] == 1
        scores_out = self.nonconformity_scores(uncal_preds[has_ones])
        masked_scores_out = scores_out * inds_ones + np.nan_to_num(
            -np.inf * (1 - inds_ones).astype(float)
        )
        calibration_scores_out = np.empty(masked_scores_out.shape[0])
        for i, score_row in enumerate(masked_scores_out):
            data_mask = mask.T[i]
            masked_scores = score_row[data_mask]
            calibration_scores_out[i] = np.max(masked_scores)

        self.tout = np.quantile(calibration_scores_out, 1 - self.conformal_alpha / 2, interpolation="higher")
        self.tin = np.quantile(calibration_scores_in, self.conformal_alpha / 2, interpolation="higher")

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())  # shape(data, task)
        scores = self.nonconformity_scores(uncal_preds)

        cal_preds_in = (scores <= self.tin).astype(int)
        cal_preds_out = (scores <= self.tout).astype(int)
        cal_preds = np.concatenate((cal_preds_in, cal_preds_out), axis=1)

        return uncal_preds.tolist(), cal_preds.tolist()


class ConformalRegressionCalibrator(UncertaintyCalibrator):
    """
    Conformal Calibrator for regression datasets. Used for both conformal regression and conformal quantile regression.
    Outputs interval of variable size, centered around quantile outputs of model, for each datapoint. Intervals 
    should cover 1-alpha proportion of datapoints.
    As discussed in https://arxiv.org/abs/2107.07511.
    """

    @property
    def label(self):
        return f"conformal_regression_{self.conformal_alpha}_half_interval"

    def raise_argument_errors(self):
        super().raise_argument_errors()
        if self.dataset_type != "regression":
            raise ValueError(
                "Conformal Regression is only implemented for regression dataset types."
            )

    def nll(
        self,
        preds: List[List[float]],
        unc: List[List[float]],
        targets: List[List[float]],
        mask: List[List[bool]],
    ):
        raise NotImplementedError(
            "The NLL uncertainty evaluation method for regression models has not been implemented for use with the conformal regression calibration method."
            )

    def calibrate(self):
        uncal_preds = np.array(self.calibration_predictor.get_uncal_preds())  # shape(data, tasks)
        uncal_interval = np.array(self.calibration_predictor.get_uncal_output())
        targets = np.array(self.calibration_data.targets())
        mask = np.array(self.calibration_data.mask())
        num_data = uncal_interval.shape[0]
        self.num_tasks = uncal_interval.shape[1]
        if self.calibration_data.is_atom_bond_targets:
            uncal_preds = [np.concatenate(x) for x in zip(*uncal_preds)]
            uncal_interval = [np.concatenate(x) for x in zip(*uncal_interval)]
            targets = [np.concatenate(x) for x in zip(*targets)]
        else:
            uncal_preds = np.array(list(zip(*uncal_preds)))
            uncal_interval = np.array(list(zip(*uncal_interval)))
            targets = targets.astype(float)
            targets = np.array(list(zip(*targets)))

        self.qhats = []
        for i in range(self.num_tasks):
            task_mask = mask[i]
            task_targets = targets[i][task_mask]
            task_preds = uncal_preds[i][task_mask]
            task_interval = uncal_interval[i][task_mask]
            uncal_interval_lower = task_preds - task_interval
            uncal_interval_upper = task_preds + task_interval
            calibration_scores = np.maximum(
                uncal_interval_lower - task_targets, task_targets - uncal_interval_upper
            )
            q_level = np.ceil((num_data + 1) * (1 - self.conformal_alpha)) / num_data
            qhat = np.quantile(calibration_scores, q_level, interpolation='higher')
            self.qhats.append(qhat)

    def apply_calibration(self, uncal_predictor: UncertaintyPredictor):
        uncal_preds = np.array(uncal_predictor.get_uncal_preds())  # shape(data, task)
        uncal_interval = np.array(uncal_predictor.get_uncal_output())  # shape(data, task)
        cal_intervals = uncal_interval + self.qhats
        return uncal_preds, cal_intervals


def build_uncertainty_calibrator(
    calibration_method: str,
    uncertainty_method: str,
    regression_calibrator_metric: str,
    interval_percentile: int,
    calibration_data: MoleculeDataset,
    calibration_data_loader: MoleculeDataLoader,
    models: Iterator[MoleculeModel],
    scalers: Iterator[StandardScaler],
    num_models: int,
    dataset_type: str,
    loss_function: str,
    uncertainty_dropout_p: float,
    conformal_alpha: float,
    dropout_sampling_size: int,
    spectra_phase_mask: List[List[bool]],
) -> UncertaintyCalibrator:
    """
    Function that chooses the subclass of :class: `UncertaintyCalibrator`
    based on the provided arguments and returns that class.
    """
    if calibration_method is None:
        if dataset_type == "regression":
            if regression_calibrator_metric == "stdev":
                calibration_method = "zscaling"
            else:
                calibration_method = "zelikman_interval"
        if dataset_type in ["classification", "multiclass"]:
            calibration_method == "isotonic"

    supported_calibrators = {
        "zscaling": ZScalingCalibrator,
        "tscaling": TScalingCalibrator,
        "zelikman_interval": ZelikmanCalibrator,
        "mve_weighting": MVEWeightingCalibrator,
        "platt": PlattCalibrator,
        "conformal": ConformalMultilabelCalibrator
        if dataset_type == "classification"
        else ConformalMulticlassCalibrator,
        "conformal_adaptive": ConformalAdaptiveMulticlassCalibrator,
        "conformal_regression": ConformalRegressionCalibrator,
        "conformal_quantile_regression": ConformalRegressionCalibrator,
        "isotonic": IsotonicCalibrator
        if dataset_type == "classification"
        else IsotonicMulticlassCalibrator,
    }

    calibrator_class = supported_calibrators.get(calibration_method, None)

    if calibrator_class is None:
        raise NotImplementedError(
            f"Calibrator type {calibration_method} is not currently supported. Avalable options are: {list(supported_calibrators.keys())}"
        )
    else:
        calibrator = calibrator_class(
            uncertainty_method=uncertainty_method,
            calibration_method=calibration_method,
            regression_calibrator_metric=regression_calibrator_metric,
            interval_percentile=interval_percentile,
            calibration_data=calibration_data,
            calibration_data_loader=calibration_data_loader,
            models=models,
            scalers=scalers,
            num_models=num_models,
            dataset_type=dataset_type,
            loss_function=loss_function,
            uncertainty_dropout_p=uncertainty_dropout_p,
            conformal_alpha=conformal_alpha,
            dropout_sampling_size=dropout_sampling_size,
            spectra_phase_mask=spectra_phase_mask,
        )
    return calibrator



from typing import Any, List, Optional

import numpy as np


class StandardScaler:
    """A :class:`StandardScaler` normalizes the features of a dataset.

    When it is fit on a dataset, the :class:`StandardScaler` learns the mean and standard deviation across the 0th axis.
    When transforming a dataset, the :class:`StandardScaler` subtracts the means and divides by the standard deviations.
    """

    def __init__(self, means: np.ndarray = None, stds: np.ndarray = None, replace_nan_token: Any = None):
        """
        :param means: An optional 1D numpy array of precomputed means.
        :param stds: An optional 1D numpy array of precomputed standard deviations.
        :param replace_nan_token: A token to use to replace NaN entries in the features.
        """
        self.means = means
        self.stds = stds
        self.replace_nan_token = replace_nan_token

    def fit(self, X: List[List[Optional[float]]]) -> 'StandardScaler':
        """
        Learns means and standard deviations across the 0th axis of the data :code:`X`.

        :param X: A list of lists of floats (or None).
        :return: The fitted :class:`StandardScaler` (self).
        """
        X = np.array(X).astype(float)
        self.means = np.nanmean(X, axis=0)
        self.stds = np.nanstd(X, axis=0)
        self.means = np.where(np.isnan(self.means), np.zeros(self.means.shape), self.means)
        self.stds = np.where(np.isnan(self.stds), np.ones(self.stds.shape), self.stds)
        self.stds = np.where(self.stds == 0, np.ones(self.stds.shape), self.stds)

        return self

    def transform(self, X: List[List[Optional[float]]]) -> np.ndarray:
        """
        Transforms the data by subtracting the means and dividing by the standard deviations.

        :param X: A list of lists of floats (or None).
        :return: The transformed data with NaNs replaced by :code:`self.replace_nan_token`.
        """
        X = np.array(X).astype(float)
        transformed_with_nan = (X - self.means) / self.stds
        transformed_with_none = np.where(np.isnan(transformed_with_nan), self.replace_nan_token, transformed_with_nan)

        return transformed_with_none

    def inverse_transform(self, X: List[List[Optional[float]]]) -> np.ndarray:
        """
        Performs the inverse transformation by multiplying by the standard deviations and adding the means.

        :param X: A list of lists of floats.
        :return: The inverse transformed data with NaNs replaced by :code:`self.replace_nan_token`.
        """
        X = np.array(X).astype(float)
        transformed_with_nan = X * self.stds + self.means
        transformed_with_none = np.where(np.isnan(transformed_with_nan), self.replace_nan_token, transformed_with_nan)

        return transformed_with_none

class AtomBondScaler(StandardScaler):
    """A :class:`AtomBondScaler` normalizes the features of a dataset.

    When it is fit on a dataset, the :class:`AtomBondScaler` learns the mean and standard deviation across the 0th axis.
    When transforming a dataset, the :class:`AtomBondScaler` subtracts the means and divides by the standard deviations.
    """

    def __init__(self, means: np.ndarray = None, stds: np.ndarray = None, replace_nan_token: Any = None, n_atom_targets = None, n_bond_targets = None):
        super().__init__(means, stds, replace_nan_token)
        self.n_atom_targets = n_atom_targets
        self.n_bond_targets = n_bond_targets

    def fit(self, X: List[List[Optional[float]]]) -> 'AtomBondScaler':
        scalers = []
        for i in range(self.n_atom_targets):
            scaler = StandardScaler().fit(X[i])
            scalers.append(scaler)
        for i in range(self.n_bond_targets):
            scaler = StandardScaler().fit(X[i+self.n_atom_targets])
            scalers.append(scaler)

        self.means = np.array([s.means for s in scalers])
        self.stds = np.array([s.stds for s in scalers])

        return self

    def transform(self, X: List[List[Optional[float]]]) -> List[np.ndarray]:
        """
        Transforms the data by subtracting the means and dividing by the standard deviations.

        :param X: A list of lists of floats (or None).
        :return: The transformed data with NaNs replaced by :code:`self.replace_nan_token`.
        """
        transformed_results = []
        for i in range(self.n_atom_targets):
            Xi = np.array(X[i]).astype(float)
            transformed_with_nan = (Xi - self.means[i]) / self.stds[i]
            transformed_with_none = np.where(np.isnan(transformed_with_nan), self.replace_nan_token, transformed_with_nan)
            transformed_results.append(transformed_with_none.tolist())
        for i in range(self.n_bond_targets):
            Xi = np.array(X[i+self.n_atom_targets]).astype(float)
            transformed_with_nan = (Xi - self.means[i+self.n_atom_targets]) / self.stds[i+self.n_atom_targets]
            transformed_with_none = np.where(np.isnan(transformed_with_nan), self.replace_nan_token, transformed_with_nan)
            transformed_results.append(transformed_with_none.tolist())

        return transformed_results

    def inverse_transform(self, X: List[List[Optional[float]]]) -> List[np.ndarray]:
        """
        Performs the inverse transformation by multiplying by the standard deviations and adding the means.

        :param X: A list of lists of floats.
        :return: The inverse transformed data with NaNs replaced by :code:`self.replace_nan_token`.
        """
        transformed_results = []
        for i in range(self.n_atom_targets):
            Xi = np.array(X[i]).astype(float)
            transformed_with_nan = Xi * self.stds[i] + self.means[i]
            transformed_with_none = np.where(np.isnan(transformed_with_nan), self.replace_nan_token, transformed_with_nan)
            transformed_results.append(transformed_with_none.tolist())
        for i in range(self.n_bond_targets):
            Xi = np.array(X[i+self.n_atom_targets]).astype(float)
            transformed_with_nan = Xi * self.stds[i+self.n_atom_targets] + self.means[i+self.n_atom_targets]
            transformed_with_none = np.where(np.isnan(transformed_with_nan), self.replace_nan_token, transformed_with_nan)
            transformed_results.append(transformed_with_none.tolist())

        return transformed_results



from typing import List
import numpy as np

from chemprop.data import MoleculeDataset


def reshape_values(
    values: List[List[List[float]]],
    test_data: MoleculeDataset,
    natom_targets: int,
    nbond_targets: int,
) -> List[List[List[float]]]:
    """
    Reshape the input from shape (num_tasks, number of atomic/bond properties for each task, 1)
    to shape (data_size, num_tasks, number of atomic/bond properties for this data in each task).

    :param values: List of atomic/bond properties with shape
                   (num_tasks, number of atomic/bond properties for each task, 1).
    :param test_data: A :class:`~chemprop.data.MoleculeDataset` containing valid datapoints.
    :param natom_targets: The number of atomic targets.
    :param nbond_targets: The number of bond targets.
    :return: List of atomic/bond properties with shape
             (data_size, num_tasks, number of atomic/bond properties for this data in each task).
    """
    n_atoms, n_bonds = test_data.number_of_atoms, test_data.number_of_bonds
    num_atom_bond_tasks = natom_targets + nbond_targets
    reshaped_values = np.empty([len(test_data), num_atom_bond_tasks], dtype=object)

    for i in range(natom_targets):
        atom_targets = values[i].reshape(-1,)
        atom_targets = np.hsplit(atom_targets, np.cumsum(np.array(n_atoms)))[:-1]
        reshaped_values[:, i] = atom_targets

    for i in range(nbond_targets):
        bond_targets = values[i+ natom_targets].reshape(-1,)
        bond_targets = np.hsplit(bond_targets, np.cumsum(np.array(n_bonds)))[:-1]
        reshaped_values[:, i + natom_targets] = bond_targets

    return reshaped_values


def reshape_individual_preds(
    individual_preds: List[List[List[List[float]]]],
    test_data: MoleculeDataset,
    natom_targets: int,
    nbond_targets: int,
    num_models: int,
) -> List[List[List[List[float]]]]:
    """
    Reshape the input from shape (num_tasks, number of atomic/bond properties for each task, 1, num_models)
    to shape (data_size, num_tasks, num_models, number of atomic/bond properties for this data in each task).

    :param individual_preds: List of atomic/bond properties with shape
                             (num_tasks, number of atomic/bond properties for each task, 1, num_models).
    :param test_data: A :class:`~chemprop.data.MoleculeDataset` containing valid datapoints.
    :param natom_targets: The number of atomic targets.
    :param nbond_targets: The number of bond targets.
    :param num_models: Number of models.
    :return: List of atomic/bond properties with shape
             (data_size, num_tasks, num_models, number of atomic/bond properties for this data in each task).
    """
    n_atoms, n_bonds = test_data.number_of_atoms, test_data.number_of_bonds
    num_atom_bond_tasks = natom_targets + nbond_targets
    individual_values = np.empty([len(test_data), num_atom_bond_tasks], dtype=object)

    for i in range(natom_targets):
        atom_targets = individual_preds[i].T.reshape(num_models, -1)
        atom_targets = np.hsplit(atom_targets, np.cumsum(np.array(n_atoms)))[:-1]
        individual_values[:, i] = atom_targets

    for i in range(nbond_targets):
        bond_targets = individual_preds[i + natom_targets].T.reshape(num_models, -1)
        bond_targets = np.hsplit(bond_targets, np.cumsum(np.array(n_bonds)))[:-1]
        individual_values[:, i + natom_targets] = bond_targets

    return individual_values



"""Defines a number of database helper functions."""

import os
import shutil
import sqlite3
from typing import Any, Dict, List, Optional, Tuple, Union

from flask import current_app, Flask, g

from chemprop.web.app import app


DB_PATH = 'chemprop.sqlite3'


def init_app(app: Flask):
    global DB_PATH

    app.teardown_appcontext(close_db)
    DB_PATH = app.config['DB_PATH']


def init_db():
    """
    Initializes the database by running schema.sql.
    This will wipe existing tables and the corresponding files.
    """
    shutil.rmtree(app.config['DATA_FOLDER'])
    os.makedirs(app.config['DATA_FOLDER'])

    shutil.rmtree(app.config['CHECKPOINT_FOLDER'])
    os.makedirs(app.config['CHECKPOINT_FOLDER'])

    db = get_db()

    with current_app.open_resource('schema.sql') as f:
        db.executescript(f.read().decode('utf8'))


def get_db():
    """
    Connects to the database.
    Returns a database object that can be queried.
    """
    if 'db' not in g:
        g.db = sqlite3.connect(
            DB_PATH,
            detect_types=sqlite3.PARSE_DECLTYPES
        )
        g.db.row_factory = sqlite3.Row

    return g.db


def query_db(query: str, args: Tuple[Any] = (), one: bool = False) -> Union[Optional[sqlite3.Row], List[sqlite3.Row]]:
    """
    Helper function to allow for easy queries.

    :param query: The query to be executed.
    :param args: The arguments to be passed into the query.
    :param one: Whether the query should return all results or just the first.
    :return The results of the query.
    """
    cur = get_db().execute(query, args)
    rv = cur.fetchall()
    cur.close()
    return (rv[0] if rv else None) if one else rv


def close_db(e: Optional[Any] = None):
    """
    Closes the connection to the database. Called after every request.

    :param e: Error object from last call.
    """
    db = g.pop('db', None)

    if db is not None:
        db.close()


# Table Specific Functions
def get_all_users() -> Dict[int, Dict[str, Any]]:
    """
    Returns all users.

    :return A dictionary of users with their ids as keys.
    """
    rows = query_db("SELECT * FROM user")

    return {row['id']: {"username": row['username'], "preferences": row['preferences']} for row in rows} if rows else {}


def insert_user(username: str) -> Tuple[int, str]:
    """
    Inserts a new user. If the desired username is already taken,
    appends integers incrementally until an open name is found.

    :param username: The desired username for the new user.
    :return A tuple containing the id and name of the new user.
    """
    db = get_db()

    new_user_id = None
    count = 0
    while new_user_id is None:
        temp_name = username

        if count != 0:
            temp_name += str(count)
        try:
            cur = db.execute('INSERT INTO user (username) VALUES (?)', [temp_name])
            new_user_id = cur.lastrowid
        except sqlite3.IntegrityError:
            count += 1

    db.commit()
    cur.close()

    return new_user_id, temp_name


def get_ckpts(user_id: int) -> List[sqlite3.Row]:
    """
    Returns the checkpoints associated with the given user.
    If no user_id is provided, return the checkpoints associated
    with the default user.

    :param user_id: The id of the user whose checkpoints are returned.
    :return A list of checkpoints.
    """
    if not user_id:
        user_id = app.config['DEFAULT_USER_ID']

    return query_db(f'SELECT * FROM ckpt WHERE associated_user = {user_id}')


def insert_ckpt(ckpt_name: str,
                associated_user: str,
                model_class: str,
                num_epochs: int,
                ensemble_size: int,
                training_size: int) -> Tuple[int, str]:
    """
    Inserts a new checkpoint. If the desired name is already taken,
    appends integers incrementally until an open name is found.

    :param ckpt_name: The desired name for the new checkpoint.
    :param associated_user: The user that should be associated with the new checkpoint.
    :param model_class: The class of the new checkpoint.
    :param num_epochs: The number of epochs the new checkpoint will run for.
    :param ensemble_size: The number of models included in the ensemble.
    :param training_size: The number of molecules used for training.
    :return A tuple containing the id and name of the new checkpoint.
    """
    db = get_db()

    new_ckpt_id = None
    count = 0
    while new_ckpt_id is None:
        temp_name = ckpt_name

        if count != 0:
            temp_name += str(count)
        try:
            cur = db.execute('INSERT INTO ckpt '
                             '(ckpt_name, associated_user, class, epochs, ensemble_size, training_size) '
                             'VALUES (?, ?, ?, ?, ?, ?)',
                             [temp_name, associated_user, model_class, num_epochs, ensemble_size, training_size])
            new_ckpt_id = cur.lastrowid
        except sqlite3.IntegrityError:
            count += 1
            continue

    db.commit()
    cur.close()

    return new_ckpt_id, temp_name


def delete_ckpt(ckpt_id: int):
    """
    Removes the checkpoint with the specified id from the database,
    associated model columns, and the corresponding files.

    :param ckpt_id: The id of the checkpoint to be deleted.
    """
    rows = query_db(f'SELECT * FROM model WHERE associated_ckpt = {ckpt_id}')

    for row in rows:
        os.remove(os.path.join(app.config['CHECKPOINT_FOLDER'], f'{row["id"]}.pt'))

    db = get_db()
    cur = db.cursor()
    db.execute(f'DELETE FROM ckpt WHERE id = {ckpt_id}')
    db.execute(f'DELETE FROM model WHERE associated_ckpt = {ckpt_id}')
    db.commit()
    cur.close()


def get_models(ckpt_id: int) -> List[sqlite3.Row]:
    """
    Returns the models associated with the given ckpt.

    :param ckpt_id: The id of the ckpt whose component models are returned.
    :return A list of models.
    """
    return query_db(f'SELECT * FROM model WHERE associated_ckpt = {ckpt_id}')


def insert_model(ckpt_id: int) -> str:
    """
    Inserts a new model.

    :param ckpt_id: The id of the checkpoint this model should be associated with.
    :return: The id of the new model.
    """
    db = get_db()
    cur = db.execute('INSERT INTO model (associated_ckpt) VALUES (?)', [ckpt_id])
    new_model_id = cur.lastrowid
    db.commit()
    cur.close()

    return new_model_id


def get_datasets(user_id: int) -> List[sqlite3.Row]:
    """
    Returns the datasets associated with the given user.
    If no user_id is provided, return the datasets associated
    with the default user.

    :param user_id: The id of the user whose datasets are returned.
    :return A list of datasets.
    """
    if not user_id:
        user_id = app.config['DEFAULT_USER_ID']

    return query_db(f'SELECT * FROM dataset WHERE associated_user = {user_id}')


def insert_dataset(dataset_name: str,
                   associated_user: str,
                   dataset_class: str) -> Tuple[int, str]:
    """
    Inserts a new dataset. If the desired name is already taken,
    appends integers incrementally until an open name is found.

    :param dataset_name: The desired name for the new dataset.
    :param associated_user: The user to be associated with the new dataset.
    :param dataset_class: The class of the new dataset.
    :return A tuple containing the id and name of the new dataset.
    """
    db = get_db()

    new_dataset_id = None
    count = 0
    while new_dataset_id is None:
        temp_name = dataset_name

        if count != 0:
            temp_name += str(count)
        try:
            cur = db.execute('INSERT INTO dataset (dataset_name, associated_user, class) VALUES (?, ?, ?)',
                             [temp_name, associated_user, dataset_class])
            new_dataset_id = cur.lastrowid
        except sqlite3.IntegrityError:
            count += 1
            continue

    db.commit()
    cur.close()

    return new_dataset_id, temp_name


def delete_dataset(dataset_id: int):
    """
    Removes the dataset with the specified id from the database,
    and deletes the corresponding file.

    :param dataset_id: The id of the dataset to be deleted.
    """
    db = get_db()
    cur = db.execute(f'DELETE FROM dataset WHERE id = {dataset_id}')
    db.commit()
    cur.close()



from typing import List

import numpy as np
import torch
from tqdm import tqdm

from chemprop.data import MoleculeDataLoader, MoleculeDataset, StandardScaler, AtomBondScaler
from chemprop.models import MoleculeModel
from chemprop.nn_utils import activate_dropout


def predict(
    model: MoleculeModel,
    data_loader: MoleculeDataLoader,
    disable_progress_bar: bool = False,
    scaler: StandardScaler = None,
    atom_bond_scaler: AtomBondScaler = None,
    return_unc_parameters: bool = False,
    dropout_prob: float = 0.0,
) -> List[List[float]]:
    """
    Makes predictions on a dataset using an ensemble of models.

    :param model: A :class:`~chemprop.models.model.MoleculeModel`.
    :param data_loader: A :class:`~chemprop.data.data.MoleculeDataLoader`.
    :param disable_progress_bar: Whether to disable the progress bar.
    :param scaler: A :class:`~chemprop.features.scaler.StandardScaler` object fit on the training targets.
    :param atom_bond_scaler: A :class:`~chemprop.data.scaler.AtomBondScaler` fitted on the atomic/bond targets.
    :param return_unc_parameters: A bool indicating whether additional uncertainty parameters would be returned alongside the mean predictions.
    :param dropout_prob: For use during uncertainty prediction only. The propout probability used in generating a dropout ensemble.
    :return: A list of lists of predictions. The outer list is molecules while the inner list is tasks. If returning uncertainty parameters as well,
        it is a tuple of lists of lists, of a length depending on how many uncertainty parameters are appropriate for the loss function.
    """
    model.eval()

    # Activate dropout layers to work during inference for uncertainty estimation
    if dropout_prob > 0.0:

        def activate_dropout_(model):
            return activate_dropout(model, dropout_prob)

        model.apply(activate_dropout_)

    preds = []

    var, lambdas, alphas, betas = [], [], [], []  # only used if returning uncertainty parameters

    for batch in tqdm(data_loader, disable=disable_progress_bar, leave=False):
        # Prepare batch
        batch: MoleculeDataset
        mol_batch = batch.batch_graph()
        features_batch = batch.features()
        atom_descriptors_batch = batch.atom_descriptors()
        atom_features_batch = batch.atom_features()
        bond_descriptors_batch = batch.bond_descriptors()
        bond_features_batch = batch.bond_features()
        constraints_batch = batch.constraints()

        if model.is_atom_bond_targets:
            natoms, nbonds = batch.number_of_atoms, batch.number_of_bonds
            natoms, nbonds = np.array(natoms).flatten(), np.array(nbonds).flatten()
            constraints_batch = np.transpose(constraints_batch).tolist()
            device = next(model.parameters()).device

            # If the path to constraints is not given, the constraints matrix needs to be reformatted.
            if constraints_batch == []:
                for _ in batch._data:
                    natom_targets = len(model.atom_targets)
                    nbond_targets = len(model.bond_targets)
                    ntargets = natom_targets + nbond_targets
                    constraints_batch.append([None] * ntargets)

            ind = 0
            for i in range(len(model.atom_targets)):
                if not model.atom_constraints[i]:
                    constraints_batch[ind] = None
                else:
                    mean, std = atom_bond_scaler.means[ind][0], atom_bond_scaler.stds[ind][0]
                    for j, natom in enumerate(natoms):
                        constraints_batch[ind][j] = (constraints_batch[ind][j] - natom * mean) / std
                    constraints_batch[ind] = torch.tensor(constraints_batch[ind]).to(device)
                ind += 1
            for i in range(len(model.bond_targets)):
                if not model.bond_constraints[i]:
                    constraints_batch[ind] = None
                else:
                    mean, std = atom_bond_scaler.means[ind][0], atom_bond_scaler.stds[ind][0]
                    for j, nbond in enumerate(nbonds):
                        constraints_batch[ind][j] = (constraints_batch[ind][j] - nbond * mean) / std
                    constraints_batch[ind] = torch.tensor(constraints_batch[ind]).to(device)
                ind += 1
            bond_types_batch = []
            for i in range(len(model.atom_targets)):
                bond_types_batch.append(None)
            for i in range(len(model.bond_targets)):
                if model.adding_bond_types and atom_bond_scaler is not None:
                    mean, std = atom_bond_scaler.means[i+len(model.atom_targets)][0], atom_bond_scaler.stds[i+len(model.atom_targets)][0]
                    bond_types = [(b.GetBondTypeAsDouble() - mean) / std for d in batch for b in d.mol[0].GetBonds()]
                    bond_types = torch.FloatTensor(bond_types).to(device)
                    bond_types_batch.append(bond_types)
                else:
                    bond_types_batch.append(None)
        else:
            bond_types_batch = None

        # Make predictions
        with torch.no_grad():
            batch_preds = model(
                mol_batch,
                features_batch,
                atom_descriptors_batch,
                atom_features_batch,
                bond_descriptors_batch,
                bond_features_batch,
                constraints_batch,
                bond_types_batch,
            )

        if model.is_atom_bond_targets:
            batch_preds = [x.data.cpu().numpy() for x in batch_preds]
            batch_vars, batch_lambdas, batch_alphas, batch_betas = [], [], [], []

            for i, batch_pred in enumerate(batch_preds):
                if model.loss_function == "mve":
                    batch_pred, batch_var = np.split(batch_pred, 2, axis=1)
                    batch_vars.append(batch_var)
                elif model.loss_function == "dirichlet":
                    if model.classification:
                        batch_alpha = np.reshape(
                            batch_pred,
                            [batch_pred.shape[0], batch_pred.shape[1] // 2, 2],
                        )
                        batch_pred = batch_alpha[:, :, 1] / np.sum(
                            batch_alpha, axis=2
                        )  # shape(data, tasks, 2)
                        batch_alphas.append(batch_alpha)
                    elif model.multiclass:
                        raise ValueError(
                            f"In atomic/bond properties prediction, {model.multiclass} is not supported."
                        )
                elif model.loss_function == "evidential":  # regression
                    batch_pred, batch_lambda, batch_alpha, batch_beta = np.split(
                        batch_pred, 4, axis=1
                    )
                    batch_alphas.append(batch_alpha)
                    batch_lambdas.append(batch_lambda)
                    batch_betas.append(batch_beta)
                batch_preds[i] = batch_pred

            # Inverse scale for each atom/bond target if regression
            if atom_bond_scaler is not None:
                batch_preds = atom_bond_scaler.inverse_transform(batch_preds)
                for i, stds in enumerate(atom_bond_scaler.stds):
                    if model.loss_function == "mve":
                        batch_vars[i] = batch_vars[i] * stds ** 2
                    elif model.loss_function == "evidential":
                        batch_betas[i] = batch_betas[i] * stds ** 2

            # Collect vectors
            preds.append(batch_preds)
            if model.loss_function == "mve":
                var.append(batch_vars)
            elif model.loss_function == "dirichlet":
                alphas.append(batch_alphas)
            elif model.loss_function == "evidential":  # regression
                lambdas.append(batch_lambdas)
                alphas.append(batch_alphas)
                betas.append(batch_betas)
        else:
            batch_preds = batch_preds.data.cpu().numpy()

            if model.loss_function == "mve":
                batch_preds, batch_var = np.split(batch_preds, 2, axis=1)
            elif model.loss_function == "dirichlet":
                if model.classification:
                    batch_alphas = np.reshape(
                        batch_preds,
                        [batch_preds.shape[0], batch_preds.shape[1] // 2, 2],
                    )
                    batch_preds = batch_alphas[:, :, 1] / np.sum(
                        batch_alphas, axis=2
                    )  # shape(data, tasks, 2)
                elif model.multiclass:
                    batch_alphas = batch_preds
                    batch_preds = batch_preds / np.sum(
                        batch_alphas, axis=2, keepdims=True
                    )  # shape(data, tasks, num_classes)
            elif model.loss_function == "evidential":  # regression
                batch_preds, batch_lambdas, batch_alphas, batch_betas = np.split(
                    batch_preds, 4, axis=1
                )

            # Inverse scale if regression
            if scaler is not None:
                batch_preds = scaler.inverse_transform(batch_preds)
                if model.loss_function == "mve":
                    batch_var = batch_var * scaler.stds**2
                elif model.loss_function == "evidential":
                    batch_betas = batch_betas * scaler.stds**2

            # Collect vectors
            batch_preds = batch_preds.tolist()
            preds.extend(batch_preds)
            if model.loss_function == "mve":
                var.extend(batch_var.tolist())
            elif model.loss_function == "dirichlet":
                alphas.extend(batch_alphas.tolist())
            elif model.loss_function == "evidential":  # regression
                lambdas.extend(batch_lambdas.tolist())
                alphas.extend(batch_alphas.tolist())
                betas.extend(batch_betas.tolist())

    if model.is_atom_bond_targets:
        preds = [np.concatenate(x) for x in zip(*preds)]
        var = [np.concatenate(x) for x in zip(*var)]
        alphas = [np.concatenate(x) for x in zip(*alphas)]
        betas = [np.concatenate(x) for x in zip(*betas)]
        lambdas = [np.concatenate(x) for x in zip(*lambdas)]

    if return_unc_parameters:
        if model.loss_function == "mve":
            return preds, var
        elif model.loss_function == "dirichlet":
            return preds, alphas
        elif model.loss_function == "evidential":
            return preds, lambdas, alphas, betas

    return preds



import json
from logging import Logger
import os
from typing import Dict, List

import numpy as np
import warnings
warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
import pandas as pd
from tensorboardX import SummaryWriter
import torch
from tqdm import trange
from torch.optim.lr_scheduler import ExponentialLR

from .evaluate import evaluate, evaluate_predictions
from .predict import predict
from .train import train
from .loss_functions import get_loss_func
from chemprop.spectra_utils import normalize_spectra, load_phase_mask
from chemprop.args import TrainArgs
from chemprop.constants import MODEL_FILE_NAME
from chemprop.data import get_class_sizes, get_data, MoleculeDataLoader, MoleculeDataset, set_cache_graph, split_data
from chemprop.models import MoleculeModel
from chemprop.nn_utils import param_count, param_count_all
from chemprop.utils import build_optimizer, build_lr_scheduler, load_checkpoint, makedirs, \
    save_checkpoint, save_smiles_splits, load_frzn_model, multitask_mean


def run_training(args: TrainArgs,
                 data: MoleculeDataset,
                 logger: Logger = None) -> Dict[str, List[float]]:
    """
    Loads data, trains a Chemprop model, and returns test scores for the model checkpoint with the highest validation score.

    :param args: A :class:`~chemprop.args.TrainArgs` object containing arguments for
                 loading data and training the Chemprop model.
    :param data: A :class:`~chemprop.data.MoleculeDataset` containing the data.
    :param logger: A logger to record output.
    :return: A dictionary mapping each metric in :code:`args.metrics` to a list of values for each task.

    """
    if logger is not None:
        debug, info = logger.debug, logger.info
    else:
        debug = info = print

    # Set pytorch seed for random initial weights
    torch.manual_seed(args.pytorch_seed)

    # Split data
    debug(f'Splitting data with seed {args.seed}')
    if args.separate_test_path:
        test_data = get_data(path=args.separate_test_path,
                             args=args,
                             features_path=args.separate_test_features_path,
                             atom_descriptors_path=args.separate_test_atom_descriptors_path,
                             bond_descriptors_path=args.separate_test_bond_descriptors_path,
                             phase_features_path=args.separate_test_phase_features_path,
                             constraints_path=args.separate_test_constraints_path,
                             smiles_columns=args.smiles_columns,
                             loss_function=args.loss_function,
                             logger=logger)
    if args.separate_val_path:
        val_data = get_data(path=args.separate_val_path,
                            args=args,
                            features_path=args.separate_val_features_path,
                            atom_descriptors_path=args.separate_val_atom_descriptors_path,
                            bond_descriptors_path=args.separate_val_bond_descriptors_path,
                            phase_features_path=args.separate_val_phase_features_path,
                            constraints_path=args.separate_val_constraints_path,
                            smiles_columns=args.smiles_columns,
                            loss_function=args.loss_function,
                            logger=logger)

    if args.separate_val_path and args.separate_test_path:
        train_data = data
    elif args.separate_val_path:
        train_data, _, test_data = split_data(data=data,
                                              split_type=args.split_type,
                                              sizes=args.split_sizes,
                                              key_molecule_index=args.split_key_molecule,
                                              seed=args.seed,
                                              num_folds=args.num_folds,
                                              args=args,
                                              logger=logger)
    elif args.separate_test_path:
        train_data, val_data, _ = split_data(data=data,
                                             split_type=args.split_type,
                                             sizes=args.split_sizes,
                                             key_molecule_index=args.split_key_molecule,
                                             seed=args.seed,
                                             num_folds=args.num_folds,
                                             args=args,
                                             logger=logger)
    else:
        train_data, val_data, test_data = split_data(data=data,
                                                     split_type=args.split_type,
                                                     sizes=args.split_sizes,
                                                     key_molecule_index=args.split_key_molecule,
                                                     seed=args.seed,
                                                     num_folds=args.num_folds,
                                                     args=args,
                                                     logger=logger)

    if args.dataset_type == 'classification':
        class_sizes = get_class_sizes(data)
        debug('Class sizes')
        for i, task_class_sizes in enumerate(class_sizes):
            debug(f'{args.task_names[i]} '
                  f'{", ".join(f"{cls}: {size * 100:.2f}%" for cls, size in enumerate(task_class_sizes))}')
        train_class_sizes = get_class_sizes(train_data, proportion=False)
        args.train_class_sizes = train_class_sizes

    if args.save_smiles_splits:
        save_smiles_splits(
            data_path=args.data_path,
            save_dir=args.save_dir,
            task_names=args.task_names,
            features_path=args.features_path,
            constraints_path=args.constraints_path,
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            smiles_columns=args.smiles_columns,
            loss_function=args.loss_function,
            logger=logger,
        )

    if args.features_scaling:
        features_scaler = train_data.normalize_features(replace_nan_token=0)
        val_data.normalize_features(features_scaler)
        test_data.normalize_features(features_scaler)
    else:
        features_scaler = None

    if args.atom_descriptor_scaling and args.atom_descriptors is not None:
        atom_descriptor_scaler = train_data.normalize_features(replace_nan_token=0, scale_atom_descriptors=True)
        val_data.normalize_features(atom_descriptor_scaler, scale_atom_descriptors=True)
        test_data.normalize_features(atom_descriptor_scaler, scale_atom_descriptors=True)
    else:
        atom_descriptor_scaler = None

    if args.bond_descriptor_scaling and args.bond_descriptors is not None:
        bond_descriptor_scaler = train_data.normalize_features(replace_nan_token=0, scale_bond_descriptors=True)
        val_data.normalize_features(bond_descriptor_scaler, scale_bond_descriptors=True)
        test_data.normalize_features(bond_descriptor_scaler, scale_bond_descriptors=True)
    else:
        bond_descriptor_scaler = None

    args.train_data_size = len(train_data)

    debug(f'Total size = {len(data):,} | '
          f'train size = {len(train_data):,} | val size = {len(val_data):,} | test size = {len(test_data):,}')

    if len(val_data) == 0:
        raise ValueError('The validation data split is empty. During normal chemprop training (non-sklearn functions), \
            a validation set is required to conduct early stopping according to the selected evaluation metric. This \
            may have occurred because validation data provided with `--separate_val_path` was empty or contained only invalid molecules.')

    if len(test_data) == 0:
        debug('The test data split is empty. This may be either because splitting with no test set was selected, \
            such as with `cv-no-test`, or because test data provided with `--separate_test_path` was empty or contained only invalid molecules. \
            Performance on the test set will not be evaluated and metric scores will return `nan` for each task.')
        empty_test_set = True
    else:
        empty_test_set = False


    # Initialize scaler and scale training targets by subtracting mean and dividing standard deviation (regression only)
    if args.dataset_type == 'regression':
        debug('Fitting scaler')
        if args.is_atom_bond_targets:
            scaler = None
            atom_bond_scaler = train_data.normalize_atom_bond_targets()
        else:
            scaler = train_data.normalize_targets()
            atom_bond_scaler = None
        args.spectra_phase_mask = None
    elif args.dataset_type == 'spectra':
        debug('Normalizing spectra and excluding spectra regions based on phase')
        args.spectra_phase_mask = load_phase_mask(args.spectra_phase_mask_path)
        for dataset in [train_data, test_data, val_data]:
            data_targets = normalize_spectra(
                spectra=dataset.targets(),
                phase_features=dataset.phase_features(),
                phase_mask=args.spectra_phase_mask,
                excluded_sub_value=None,
                threshold=args.spectra_target_floor,
            )
            dataset.set_targets(data_targets)
        scaler = None
        atom_bond_scaler = None
    else:
        args.spectra_phase_mask = None
        scaler = None
        atom_bond_scaler = None

    # Get loss function
    loss_func = get_loss_func(args)

    # Set up test set evaluation
    test_smiles, test_targets = test_data.smiles(), test_data.targets()
    if args.dataset_type == 'multiclass':
        sum_test_preds = np.zeros((len(test_smiles), args.num_tasks, args.multiclass_num_classes))
    elif args.is_atom_bond_targets:
        sum_test_preds = []
        for tb in zip(*test_data.targets()):
            tb = np.concatenate(tb)
            sum_test_preds.append(np.zeros((tb.shape[0], 1)))
        sum_test_preds = np.array(sum_test_preds, dtype=object)
    else:
        sum_test_preds = np.zeros((len(test_smiles), args.num_tasks))

    # Automatically determine whether to cache
    if len(data) <= args.cache_cutoff:
        set_cache_graph(True)
        num_workers = 0
    else:
        set_cache_graph(False)
        num_workers = args.num_workers

    # Create data loaders
    train_data_loader = MoleculeDataLoader(
        dataset=train_data,
        batch_size=args.batch_size,
        num_workers=num_workers,
        class_balance=args.class_balance,
        shuffle=True,
        seed=args.seed
    )
    val_data_loader = MoleculeDataLoader(
        dataset=val_data,
        batch_size=args.batch_size,
        num_workers=num_workers
    )
    test_data_loader = MoleculeDataLoader(
        dataset=test_data,
        batch_size=args.batch_size,
        num_workers=num_workers
    )

    if args.class_balance:
        debug(f'With class_balance, effective train size = {train_data_loader.iter_size:,}')

    # Train ensemble of models
    for model_idx in range(args.ensemble_size):
        # Tensorboard writer
        save_dir = os.path.join(args.save_dir, f'model_{model_idx}')
        makedirs(save_dir)
        try:
            writer = SummaryWriter(log_dir=save_dir)
        except:
            writer = SummaryWriter(logdir=save_dir)

        # Load/build model
        if args.checkpoint_paths is not None:
            debug(f'Loading model {model_idx} from {args.checkpoint_paths[model_idx]}')
            model = load_checkpoint(args.checkpoint_paths[model_idx], logger=logger)
        else:
            debug(f'Building model {model_idx}')
            model = MoleculeModel(args)

        # Optionally, overwrite weights:
        if args.checkpoint_frzn is not None:
            debug(f'Loading and freezing parameters from {args.checkpoint_frzn}.')
            model = load_frzn_model(model=model, path=args.checkpoint_frzn, current_args=args, logger=logger)

        debug(model)

        if args.checkpoint_frzn is not None:
            debug(f'Number of unfrozen parameters = {param_count(model):,}')
            debug(f'Total number of parameters = {param_count_all(model):,}')
        else:
            debug(f'Number of parameters = {param_count_all(model):,}')

        if args.cuda:
            debug('Moving model to cuda')
        model = model.to(args.device)

        # Ensure that model is saved in correct location for evaluation if 0 epochs
        save_checkpoint(os.path.join(save_dir, MODEL_FILE_NAME), model, scaler,
                        features_scaler, atom_descriptor_scaler, bond_descriptor_scaler,
                        atom_bond_scaler, args)

        # Optimizers
        optimizer = build_optimizer(model, args)

        # Learning rate schedulers
        scheduler = build_lr_scheduler(optimizer, args)

        # Run training
        best_score = float('inf') if args.minimize_score else -float('inf')
        best_epoch, n_iter = 0, 0
        for epoch in trange(args.epochs):
            debug(f'Epoch {epoch}')
            n_iter = train(
                model=model,
                data_loader=train_data_loader,
                loss_func=loss_func,
                optimizer=optimizer,
                scheduler=scheduler,
                args=args,
                n_iter=n_iter,
                atom_bond_scaler=atom_bond_scaler,
                logger=logger,
                writer=writer
            )
            if isinstance(scheduler, ExponentialLR):
                scheduler.step()
            val_scores = evaluate(
                model=model,
                data_loader=val_data_loader,
                num_tasks=args.num_tasks,
                metrics=args.metrics,
                dataset_type=args.dataset_type,
                scaler=scaler,
                quantiles=args.quantiles,
                atom_bond_scaler=atom_bond_scaler,
                logger=logger
            )

            for metric, scores in val_scores.items():
                # Average validation score\
                mean_val_score = multitask_mean(
                    scores=scores,
                    metric=metric,
                    ignore_nan_metrics=args.ignore_nan_metrics
                )
                debug(f'Validation {metric} = {mean_val_score:.6f}')
                writer.add_scalar(f'validation_{metric}', mean_val_score, n_iter)

                if args.show_individual_scores:
                    if args.loss_function == "quantile_interval" and metric == "quantile":
                        num_tasks = len(args.task_names) // 2
                        task_names = args.task_names[:num_tasks]
                        task_names = [f"{task_name} lower" for task_name in task_names] + [
                                        f"{task_name} upper" for task_name in task_names]
                    else:
                        task_names = args.task_names
                    # Individual validation scores
                    for task_name, val_score in zip(task_names, scores):
                        debug(f'Validation {task_name} {metric} = {val_score:.6f}')
                        writer.add_scalar(f'validation_{task_name}_{metric}', val_score, n_iter)

            # Save model checkpoint if improved validation score
            mean_val_score = multitask_mean(
                scores=val_scores[args.metric],
                metric=args.metric,
                ignore_nan_metrics=args.ignore_nan_metrics
            )
            if args.minimize_score and mean_val_score < best_score or \
                    not args.minimize_score and mean_val_score > best_score:
                best_score, best_epoch = mean_val_score, epoch
                save_checkpoint(os.path.join(save_dir, MODEL_FILE_NAME), model, scaler, features_scaler,
                                atom_descriptor_scaler, bond_descriptor_scaler, atom_bond_scaler, args)

        # Evaluate on test set using model with best validation score
        info(f'Model {model_idx} best validation {args.metric} = {best_score:.6f} on epoch {best_epoch}')
        model = load_checkpoint(os.path.join(save_dir, MODEL_FILE_NAME), device=args.device, logger=logger)

        if empty_test_set:
            info(f'Model {model_idx} provided with no test set, no metric evaluation will be performed.')
        else:
            test_preds = predict(
                model=model,
                data_loader=test_data_loader,
                scaler=scaler,
                atom_bond_scaler=atom_bond_scaler
            )
            test_scores = evaluate_predictions(
                preds=test_preds,
                targets=test_targets,
                num_tasks=args.num_tasks,
                metrics=args.metrics,
                dataset_type=args.dataset_type,
                is_atom_bond_targets=args.is_atom_bond_targets,
                gt_targets=test_data.gt_targets(),
                lt_targets=test_data.lt_targets(),
                quantiles=args.quantiles,
                logger=logger
            )

            if len(test_preds) != 0:
                if args.is_atom_bond_targets:
                    sum_test_preds += np.array(test_preds, dtype=object)
                else:
                    sum_test_preds += np.array(test_preds)

            # Average test score
            for metric, scores in test_scores.items():
                avg_test_score = np.nanmean(scores)
                info(f'Model {model_idx} test {metric} = {avg_test_score:.6f}')
                writer.add_scalar(f'test_{metric}', avg_test_score, 0)

                if args.show_individual_scores and args.dataset_type != 'spectra':
                    # Individual test scores
                    for task_name, test_score in zip(task_names, scores):
                        info(f'Model {model_idx} test {task_name} {metric} = {test_score:.6f}')
                        writer.add_scalar(f'test_{task_name}_{metric}', test_score, n_iter)
        writer.close()

    # Evaluate ensemble on test set
    if empty_test_set:
        ensemble_scores = {
            metric: [np.nan for task in args.task_names] for metric in args.metrics
        }
    else:
        avg_test_preds = (sum_test_preds / args.ensemble_size).tolist()

        ensemble_scores = evaluate_predictions(
            preds=avg_test_preds,
            targets=test_targets,
            num_tasks=args.num_tasks,
            metrics=args.metrics,
            dataset_type=args.dataset_type,
            is_atom_bond_targets=args.is_atom_bond_targets,
            gt_targets=test_data.gt_targets(),
            lt_targets=test_data.lt_targets(),
            quantiles=args.quantiles,
            logger=logger,
        )

    for metric, scores in ensemble_scores.items():
        # Average ensemble score
        mean_ensemble_test_score = multitask_mean(
            scores=scores,
            metric=metric,
            ignore_nan_metrics=args.ignore_nan_metrics
        )
        info(f'Ensemble test {metric} = {mean_ensemble_test_score:.6f}')

        # Individual ensemble scores
        if args.show_individual_scores:
            for task_name, ensemble_score in zip(task_names, scores):
                info(f'Ensemble test {task_name} {metric} = {ensemble_score:.6f}')

    # Save scores
    with open(os.path.join(args.save_dir, 'test_scores.json'), 'w') as f:
        json.dump(ensemble_scores, f, indent=4, sort_keys=True)

    # Optionally save test preds
    if args.save_preds and not empty_test_set:
        test_preds_dataframe = pd.DataFrame(data={'smiles': test_data.smiles()})

        if args.is_atom_bond_targets:
            n_atoms, n_bonds = test_data.number_of_atoms, test_data.number_of_bonds

            for i, atom_target in enumerate(args.atom_targets):
                values = np.split(np.array(avg_test_preds[i]).flatten(), np.cumsum(np.array(n_atoms)))[:-1]
                values = [list(v) for v in values]
                test_preds_dataframe[atom_target] = values
            for i, bond_target in enumerate(args.bond_targets):
                values = np.split(np.array(avg_test_preds[i+len(args.atom_targets)]).flatten(), np.cumsum(np.array(n_bonds)))[:-1]
                values = [list(v) for v in values]
                test_preds_dataframe[bond_target] = values
        else:
            if args.loss_function == "quantile_interval" and metric == "quantile":
                num_tasks = len(args.task_names) // 2
                task_names = args.task_names[:num_tasks]
                avg_test_preds = np.array(avg_test_preds)
                num_data = avg_test_preds.shape[0]
                preds = avg_test_preds.reshape(num_data, 2, num_tasks).mean(axis=1)
                intervals = abs(np.diff(avg_test_preds.reshape(num_data, 2, num_tasks), axis=1) / 2)
                intervals = intervals.reshape(num_data, num_tasks)
                for i, task_name in enumerate(task_names):
                    test_preds_dataframe[task_name] = [pred[i] for pred in preds]
                for i, task_name in enumerate(task_names):
                    task_name = f"{task_name}_{args.quantile_loss_alpha}_half_interval"
                    test_preds_dataframe[task_name] = [interval[i] for interval in intervals]
            else:
                for i, task_name in enumerate(task_names):
                    test_preds_dataframe[task_name] = [pred[i] for pred in avg_test_preds]

        test_preds_dataframe.to_csv(os.path.join(args.save_dir, 'test_preds.csv'), index=False)

    return ensemble_scores
